import argparse
import logging
from logging.config import dictConfig
import os
import sys
import subprocess
import inspect
from collections import OrderedDict

import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))

from lightningcast.utils import init_config_env, extract_keypath_values
from lightningcast.logging_def import *

import lightningcast.logging_def as ld


def new_logger(logfile):
    # Set up logging
    logDict = ld.logging_def(logfile=logfile)
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = ld.StreamToLogger(stderr_logger, logging.ERROR)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="Takes a list of datetimes or rabbitMQ real-time listing of GOES scans and \
                      will make predictions of lightning in the next 60 min. Can make images, jsons, or netCDFs."
    )

    parser.add_argument(
        "input_file",
        help="Either a file to process or to a filelist containing file(s) to process.",
        type=str,
    )

    parser.add_argument(
        "--yaml_config",
        help="Yaml files to use to override default config. Will be a yaml cascade where files listed later override earlier files.",
        type=str,
        dest="yaml_files",
        default=[],
        nargs="+",
    )
    parser.add_argument(
        "--yaml_overrides",
        nargs="+",
        help="compound.key=value YAML expression as final configuration",
    )

    args = parser.parse_args()

    lc_base = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    if not os.path.islink(f"{lc_base}/lightningcast"):
        subprocess.run(["/bin/bash", "-c", f"cd {lc_base} && ln -s . lightningcast"])

    default_config_yamls = [
        os.path.join(root, file)
        for root, dirs, files in os.walk(os.path.abspath(f"{lc_base}/config/default/"))
        for file in files
    ]
    default_config_yamls.sort()
    yaml_files = default_config_yamls + args.yaml_files

    if args.yaml_overrides:
        yaml_overrides = OrderedDict(extract_keypath_values(args.yaml_overrides))
    else:
        yaml_overrides = {}

    format_context = {
        "lc_base": lc_base,
        "pwd": os.getcwd(),
    }

    config = init_config_env(yaml_files, yaml_overrides, format_context)

    from lightningcast.predict_ltg import (
        predict_ltg,
    )  # This needs to be after env variables are set. When we switch to doing config differently it can maybe be moved to the top.

    logDict = logging_def(logfile=None)
    # do this regardless of realtime/offline, so we get formatted stdout messages
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    if config.get("logfile", None) is not None:
        os.makedirs(os.path.dirname(config["logfile"]), exist_ok=True)
        new_logger(config["logfile"])

    ll_bbox = (
        None
        if config["ll_bbox"] and (config["ll_bbox"][0] == -999)
        else config["ll_bbox"]
    )
    ind_bbox = (
        None
        if config["ind_bbox"] and (config["ind_bbox"][0] == -999)
        else config["ind_bbox"]
    )
    # get_path returns a path object but predict_ltg expects a string and using path objects causes and error
    model = config["model"]
    model_config_file = config["model_config_file"]

    # I reworked this abit to make more sense with the yaml setup.
    if not os.path.isfile(model_config_file):
        model_config_file = os.path.dirname(model) + "/model_config.pkl"

    assert os.path.isfile(model_config_file)
    assert os.path.isfile(model)

    predict_ltg(
        args.input_file,
        model,
        model_config_file,
        outdir=config["outdir"],
        ll_bbox=ll_bbox,
        ind_bbox=ind_bbox,
        datadir_patt=config["datadir_patt"],
        values=config["prob_values"],
        glmpatt=config["glmpatt"],
        glmvar=config["glmvar"],
        eni_path=config["eni_path"],
        mrms_patt=config["mrms_patt"],
        sector_suffix=config["sector_suffix"],
        sector=config["sector"],
        infinite=False,
        re_upload=config["re_upload"],
        make_img=config["make_img"],
        colors=config["colors"],
        contour_cmap=config["contour_cmap"],
        county_map=config["county_map"],
        lightning_meteograms=config["lightning_meteograms"],
        geotiff=config["geotiff"],
        savegrid=config["savegrid"],
        fire_event_ts=config["fire_event_ts"],
        make_json=config["make_json"],
        pickle_preds_labs=config["pickle_preds_labs"],
        ltg_stats=config["ltg_stats"],
        pixel_buff_and_stride=config["pixel_buff_and_stride"],
        timeseries=config["timeseries"],
        grplacefile=config["grplacefile"],
        plax_hgt=config["plax_hgt"],
        plax_pkl=config["plax_pkl"],
        extra_script=config["extra_script"],
        ftp=config["ftp"],
        ref10=config["ref10"],
        plot_engln=config["plot_engln"],
        plot_points=config["plot_points"],
        sat_info=config["sat_info"],
    )
