import os, sys
import numpy as np
import time
from subprocess import call, Popen
from datetime import datetime, timedelta
import shutil
from glob import glob
import re
import logging
from logging.config import dictConfig
import argparse

import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))

from lightningcast.logging_def import *
import traceback
from lightningcast import utils

pltg_data = os.environ["PLTG_DATA"] + "/"


def cleanup():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-dd",
        "--datadir",
        help="Root datadir. Default = ${PLTG_DATA}",
        default=[pltg_data],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-c",
        "--clean_days_thresh",
        help="Default = 30; Will remove directories older than this.",
        type=int,
        default=[30],
        nargs=1,
    )
    parser.add_argument(
        "-l",
        "--logfile",
        help="Full path/filename of logfile. Optional.",
        type=str,
        nargs=1,
    )

    args = parser.parse_args()

    datadir = args.datadir[0]
    if datadir[-1] != "/":
        datadir += "/"
    clean_days_thresh = args.clean_days_thresh[0]
    logfile = args.logfile[0] if (args.logfile) else None
    # define the logger
    logDict = logging_def(logfile=logfile)
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl

    # open /dev/null
    devnull = open(os.devnull, "w")

    # clean_days_thresh is basically how many days of data do you want to keep around
    # -- You often want to save off data before removing past this threshold, so keep at least 30 days.

    logging.info("Process started")

    while True:
        today = datetime.utcnow().strftime("%Y%m%d")
        yesterday = (datetime.utcnow() - timedelta(days=1)).strftime("%Y%m%d")

        NWP_DIR = f"{datadir}/NWP/"
        LTGDIR = f"{datadir}/lightning/"
        PRODUCTS_DIR = f"{datadir}/products/"
        NWP_RAW_DIR = f"{datadir}/NWP_RAW/"
        LTGDEST = "/ships19/grain/probsevere/lightning/"
        ENIDIR = f"{datadir}/lightning/realtime/ENI/"
        REMOTE_ENIDIR = "/ships19/grain/lightning/ENI_1min/"
        LUSTREDIR = "/ships19/grain/common/LightningCast/realtime/products/"

        nowsecs = int(time.time())
        # Cleaning up date directories. clean_time in seconds
        try:
            clean_days_thresh = int(clean_days_thresh)
        except ValueError as err:
            logging.error(str(err))
            sys.exit(1)
        else:
            clean_time = nowsecs - 60 * 60 * 24 * clean_days_thresh
        # make delete thresholds
        gzip_time = nowsecs - 60 * 60 * 3  # 3 hours
        move_time = nowsecs - 60 * 30 * 1  # 30 min
        delete_time = nowsecs - 60 * 30  # 30 min
        lustrepurge_time = nowsecs - 3600 * 24  # 1day

        # ==moving data
        logging.info("Deleting some GLM agg data...")
        for SAT in ["goes_east", "goes_west", "goes_west/ussamoa"]:
            for DIR in ["agg"]:
                globbed_files = glob(f"{LTGDIR}/realtime/GLM/{SAT}/{DIR}/*")
                if len(globbed_files) > 0:
                    for f in globbed_files:
                        try:
                            mtime = os.path.getmtime(f)
                        except EnvironmentError:
                            pass
                        else:
                            if mtime < delete_time:
                                try:
                                    os.remove(f)
                                except (OSError, IOError) as err:
                                    logging.error(str(err))

        # ==remove raw GLM output from glmtools script
        rtGLMDIR = LTGDIR + "realtime/GLM/"
        for SAT in ["goes_east", "goes_west", "goes_west/ussamoa"]:
            rawdir = f"{LTGDIR}/realtime/GLM/{SAT}/raw/"
            if os.path.isdir(rawdir):
                logging.info(f"Removing raw remapped GLM netcdfs in {rawdir}")
                for dirName, subdirList, fileList in os.walk(
                    rawdir, topdown=False
                ):  # walking bottom-up
                    # delete any old files
                    for fname in fileList:
                        theFile = os.path.join(dirName, fname)
                        mtime = os.path.getmtime(theFile)
                        if mtime < delete_time:
                            try:
                                os.remove(theFile)
                            except (OSError, IOError) as err:
                                logging.error(str(err))
                    # check for empty directories (i.e., no files AND no subdirs)
                    if not (fileList) and not (subdirList):
                        mtime = os.path.getmtime(dirName)
                        if mtime < delete_time:
                            try:
                                shutil.rmtree(dirName)
                            except (OSError, IOError) as err:
                                logging.error(str(err))

        # ==remove netcdfs, as they can add up quickly (especially mesos)
        logging.info(f"Removing netcdfs older than X mins...")
        for date in [today, yesterday]:
            for SAT in [
                "goes_east",
                "goes_west",
                "ahi_japan",
            ]:  # /data/PLTG/products/20220307/goes_east/RadM1/netcdf
                try:
                    sectors = next(os.walk(f"{PRODUCTS_DIR}/{date}/{SAT}/"))[
                        1
                    ]  # top-level dirs (e.g., 'RadC', 'RadM1', 'FLDK_GUAM')
                except StopIteration:  # SAT doesn't exist
                    continue
                else:
                    for sector in sectors:
                        netcdfs = glob(
                            f"{PRODUCTS_DIR}/{date}/{SAT}/{sector}/netcdf/*.nc*"
                        )
                        for nc in netcdfs:
                            mtime = os.path.getmtime(nc)
                            if mtime < delete_time:
                                try:
                                    os.remove(nc)
                                except (OSError, IOError) as err:
                                    logging.error(str(err))

        # Remove "old" ENI files
        for ff in glob(f"{ENIDIR}/*csv"):
            mtime = os.path.getmtime(ff)
            if mtime < delete_time:
                try:
                    os.remove(ff)
                except (OSError, IOError) as err:
                    logging.error(str(err))

        # Remove "old" ENI files on lustre. These are used by other users,
        # but the data are pulled in and managed by LightningCast stack.
        for ff in glob(f"{REMOTE_ENIDIR}/*csv"):
            mtime = os.path.getmtime(ff)
            if mtime < delete_time:
                try:
                    os.remove(ff)
                except (OSError, IOError) as err:
                    logging.error(str(err))

        # Purge files on remote lustre volume. These are used by other users (e.g., VOLCAT fires)
        logging.info(f"Removing old files from {LUSTREDIR}...")
        for SAT in ["goes_east", "goes_west"]:
            try:
                sectors = next(os.walk(f"{LUSTREDIR}/{SAT}/"))[
                    1
                ]  # top-level dirs (e.g., 'RadC', 'RadM1', 'FLDK_GUAM')
            except StopIteration:  # SAT doesn't exist
                continue
            else:
                for sector in sectors:
                    netcdfs = glob(f"{LUSTREDIR}/{SAT}/{sector}/*.nc*")
                    for nc in netcdfs:
                        mtime = os.path.getmtime(nc)
                        if mtime < lustrepurge_time:
                            try:
                                os.remove(nc)
                            except (OSError, IOError) as err:
                                logging.error(str(err))

        # ==gzipping NWP netcdfs
        # logging.info('gzipping netcdfs')
        # dates = glob(os.path.join(NWP_DIR,'2*'))
        # for date in dates:
        #  ncs = glob(os.path.join(date,'*.nc'))
        #  if(len(ncs) > 0):
        #    for f in ncs:
        #      mtime = os.path.getmtime(f)
        #      if(mtime < gzip_time):
        #        s = Popen(['gzip','-f',f], stderr=devnull) #==will run in background

        # ==remove raw NWP data
        # logging.info('Removing raw NWP data (HRRR grib2s) based on threshold of ' + str(clean_days_thresh) + ' days')
        # for basedir in [NWP_RAW_DIR + 'GRIB2/']:
        #  for dirName, subdirList, fileList in os.walk(basedir,topdown=False): #walking bottom-up):
        #    #files
        #    for fname in fileList:
        #      theFile = dirName + '/' + fname
        #      mtime = os.path.getmtime(theFile)
        #      if(mtime < clean_time):
        #        try:
        #          os.remove(theFile)
        #        except (OSError,IOError) as err:
        #          logging.error(str(err))

        #    #check for empty directories (i.e., no files AND no subdirs)
        #    if(not(fileList) and not(subdirList)):
        #      mtime = os.path.getmtime(dirName)
        #      if(mtime < clean_time):
        #        try:
        #          shutil.rmtree(dirName)
        #        except(OSError,IOError) as err:
        #          logging.error(str(err))

        ##################################################################
        # Now do cleanup of whole directories older than clean_days_thresh#
        ##################################################################
        if clean_days_thresh >= 1:
            logging.info(
                f"Checking for data in {datadir} older than {clean_days_thresh} days"
            )
            # NWP data
            # globbed_dirs = glob(('/').join([NWP_DIR,'2*']))
            # if(len(globbed_dirs) > 0):
            #  for d in globbed_dirs:
            #    mtime = os.path.getmtime(d)
            #    if(mtime < clean_time):
            #      logging.info('Deleting NWP data: ' + d)
            #      shutil.rmtree(d)
            # jsons, netcdfs
            globbed_dirs = glob(os.path.join(PRODUCTS_DIR, "2*"))
            if len(globbed_dirs) > 0:
                for d in globbed_dirs:
                    mtime = os.path.getmtime(d)
                    if mtime < clean_time:
                        logging.info("Deleting PRODUCT data: " + d)
                        shutil.rmtree(d)

        logging.info("sleeping.")
        time.sleep(60 * 5)
    # endwhile

    logging.info("Process ended")


if __name__ == "__main__":
    cleanup()
