import numpy as np
import xarray as xr
import pandas as pd
import os, sys
import pickle
from datetime import datetime, timedelta
import glob
from lightningcast import utils
from netCDF4 import Dataset
import shutil
import argparse
from subprocess import Popen, call, PIPE

PWD = os.environ["PWD"] + "/"
import traceback

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"  # visible in this process + all children

# CONSTANTS
# c = 2.99792458e8 #speed of light (m/s)
# h = 6.62607015e-34 #Planck's constant (J s)
# k = 1.38064852e-23 #Boltzmann's constant (J/K)
# l14 = 11.2e-6 #central wavelength of CH14 (m)

# formula for converting from radiance to BT.
# see https://www.star.nesdis.noaa.gov/goesr/docs/ATBD/Imagery.pdf
# T = [ fk2 / (alog((fk1 / Lλ) + 1)) - bc1 ] / bc2   eqn (3-5)

# for CH02 radiance to reflectance:
# see pdf above pp. 20-22
###############################################################################################################
def read_netcdf(file, data={}, global_atts={}):

    if file[-3:] == ".gz":
        gzip = 1
        unzipped_file = file[0 : len(file) - 3]
        the_file = unzipped_file
        try:
            s = call(["gunzip", "-f", file])
        except:
            print("cannot open " + file)
            return False
    else:
        gzip = 0
        the_file = file

    try:
        openFile = Dataset(the_file, "r")
    except (OSError, IOError):
        print(traceback.format_exc())
        return False

    for v in data:
        if v in openFile.variables:
            data[v] = openFile.variables[v][:]

    # get global attributes
    for gatt in global_atts:
        if gatt in openFile.ncattrs():
            global_atts[gatt] = getattr(
                openFile, gatt
            )  # global_atts is modified and returned
        else:
            print(gatt + " not in " + file)
            openFile.close()
            return False

    openFile.close()

    # --gzip file after closing?
    if gzip:
        s = Popen(
            ["gzip", "-f", unzipped_file]
        )  # ==will run in background since it doesn't need output from command.
        # s = call(['gzip','-f',unzipped_file])

    return True


###############################################################################################################
def get_glm_data(glm_file, data={}, global_atts={}, plax=None):

    if glm_file[-3:] == ".gz":
        gzip = 1
        unzipped_file = glm_file[0 : len(glm_file) - 3]
        the_file = unzipped_file
        try:
            s = call(["gunzip", "-f", glm_file])
        except:
            print("cannot open " + glm_file)
            return False
    else:
        gzip = 0
        the_file = glm_file

    try:
        openFile = Dataset(the_file, "r")
    except (OSError, IOError):
        print(traceback.format_exc())
        return False

    glmvars = [
        "flash_extent_density",
        "flash_centroid_density",
        "total_energy",
        "average_flash_area",
    ]
    for v in glmvars:
        if v in openFile.variables:
            if plax:  # this only works for 2D vars
                print("Parallax correcting " + v)
                tmpvar = openFile.variables[v][:]
                plax_var = np.zeros(np.shape(tmpvar))
                plax_var[plax] = np.ravel(tmpvar)
                data[v] = plax_var
            else:
                data[v] = openFile.variables[v][:]

    # get global attributes
    # for gatt in global_atts:
    #  if(gatt in openFile.ncattrs()):
    #    global_atts[gatt] = getattr(openFile,gatt) #global_atts is modified and returned
    #  else:
    #    print(gatt + ' not in ' + glm_file)
    #    openFile.close()
    #    return False

    openFile.close()

    # --gzip file after closing?
    if gzip:
        s = Popen(
            ["gzip", "-f", unzipped_file]
        )  # ==will run in background since it doesn't need output from command.
        # s = call(['gzip','-f',unzipped_file])

    return True


###############################################################################################################


def main():

    vars_to_write = {
        "CH02": {
            "atts": {
                "desc": "x by y slice of 0.65-micron refl from ABI",
                "units": "dimensionless",
            }
        },
        "CH05": {
            "atts": {
                "desc": "x by y slice of 1.6-micron refl from ABI",
                "units": "dimensionless",
            }
        },
        "CH06": {
            "atts": {
                "desc": "x by y slice of 2.26-micron refl from ABI",
                "units": "dimensionless",
            }
        },
        "CH13": {
            "atts": {"desc": "x by y slice of 10.3-micron BT from ABI", "units": "K"}
        },
        "FED": {
            "atts": {
                "desc": "GLM flash extent density, accumulated over 5 min",
                "units": "fl",
            }
        },
    }

    # just some placeholders for now
    global_atts = {
        "lat": -1,
        "lon": -1,
        "ABIx": -1,
        "ABIy": -1,
        "ABIx05": -1,
        "ABIy05": -1,
        "ABIx10": -1,
        "ABIy10": -1,
        "ABI_time": -1,
        "maxFED_t0": -1,
        "maxFED_t5": -1,
        "maxFED_t10": -1,
        "maxFED_t15": -1,
        "maxFED_t20": -1,
        "maxFED_t25": -1,
        "maxFED_t30": -1,
        "SST": -1,
        "mlcape": -1,
        "u925": -1,
        "v925": -1,
        "u850": -1,
        "v850": -1,
        "u700": -1,
        "v700": -1,
        "u500": -1,
        "v500": -1,
    }

    satdatadir = "/arcdata/goes/grb/goes16/"
    copy = True
    sstdir = "/iliad/grain/common/oconus_ps/OISST/"
    glmdir = "/data/jcintineo/AI/GLM_TRAIN/"
    hrrrdir = "/data/jcintineo/HRRR/remapped/"
    outbase = "/data/jcintineo/AI/OCONUS/netcdf/"

    abi_channels = ["02", "05", "06", "13"]
    glm_channels = ["FED"]
    halfsize = 15

    # get ABI lats/lons
    area_def, ny, nx, abi_lons, abi_lats = utils.get_area_definition(
        "static/GOES16.nc", return_latlons=True
    )
    # get land/sea mask
    mask_nc = Dataset("static/mask_GOES-East_CONUS_25km.nc", "r")
    landsea_mask = mask_nc.variables["mask"][
        :
    ]  # this is on the ABI grid with effective resolution of 25km.
    mask_nc.close()

    GLM_truth_buffer = 2  # in pixels
    stride = 32  # spatial stride in pixels in both x and y directions
    predict_leadtime = 30  # in min
    pred_steps = predict_leadtime / 5 + 1
    global_atts["GLM_truth_buffer"] = str(GLM_truth_buffer) + " pixels"
    global_atts["predict_leadtime"] = predict_leadtime
    global_atts["spatial_stride"] = str(stride) + " pixels"

    # with respect to ABI grid
    startX = 1510
    endX = 2499
    startY = 0
    endY = 850  # this should be north of the "Bermuda bars/strobes"

    curr_hrrr_file = ""
    curr_sst_file = ""

    data = {}

    for d in [
        20180521,
        20180615,
        20180616,
        20180331,
        20180709,
        20180710,
        20180809,
        20180818,
    ]:
        strdate = str(d)

        topdt = datetime.strptime(strdate, "%Y%m%d")

        satdir1 = topdt.strftime(satdatadir + "/%Y/%Y_%m_%d_%j/abi/L1b/RadC/")

        # sat files for the day #CHANGE THE 2* HERE ??????
        satfiles = glob.glob(
            os.path.join(satdir1, topdt.strftime("OR_ABI-L1b-RadC-M*C13_G16_s%Y%j*"))
        )
        satfiles = np.sort(satfiles)

        satdts = np.array(
            [
                datetime.strptime(
                    os.path.basename(x).split("_s")[1][0:13], "%Y%j%H%M%S"
                )
                for x in satfiles
            ]
        )

        for dt in satdts:
            cont = False
            yyyymmdd = dt.strftime("%Y%m%d")
            outts = dt.strftime("%Y%m%d-%H%M%S")
            data["FED"] = []

            # get all of the data first. Then do extraction.

            print(dt)

            # get satellite data
            for ch in abi_channels:
                netcdfpattern = dt.strftime(
                    satdatadir
                    + "/%Y/%Y_%m_%d_%j/abi/L1b/RadC/OR_ABI-L1b-RadC-M*C"
                    + ch
                    + "_G16_s%Y%j%H%M*"
                )
                try:
                    netcdf = glob.glob(netcdfpattern)[0]
                except IndexError as err:
                    # a desired channel does not exist
                    print(traceback.format_exc())
                    print("Moving on to next sat DT...")
                    continue
                # copy file from satdatadir
                try:
                    shutil.copy(netcdf, PWD)
                except (PermissionError, OSError, IOError) as err:
                    print(traceback.format_exc())
                    print("Moving on to next sat DT...")
                    continue

                nc = Dataset(PWD + os.path.basename(netcdf), "r")
                radiance = nc.variables["Rad"][:]
                if int(ch) <= 6:
                    # convert vis radiance to reflectance
                    data["CH" + ch] = radiance * nc.variables["kappa0"][:]
                else:
                    # convert IR radiance to brightness temperature
                    data["CH" + ch] = (
                        nc.variables["planck_fk2"][:]
                        / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
                        - nc.variables["planck_bc1"][:]
                    ) / nc.variables["planck_bc2"][:]

                # clean up
                nc.close()
                os.remove(PWD + os.path.basename(netcdf))

            # get GLM data
            # we can't necessarily assume sat files will be every 5 min
            tmpdt = dt
            while len(data["FED"]) < pred_steps:
                netcdfpattern = tmpdt.strftime(
                    os.path.join(
                        glmdir, "%Y%m%d", "GLM", "goes_east", "agg", "%Y%m%d-%H%M*"
                    )
                )
                try:
                    netcdf = glob.glob(netcdfpattern)[0]
                except IndexError as err:
                    print(traceback.format_exc())
                    print(netcdfpattern)
                    cont = True
                    break
                else:
                    glm_data = {}
                    new_glm_data = get_glm_data(netcdf, data=glm_data)
                    if new_glm_data:
                        data["FED"].append(glm_data["flash_extent_density"])
                    else:
                        data["FED"].append(np.zeros((1499, 2499)) - 999)  # missing grid
                    tmpdt = tmpdt + timedelta(minutes=5)
            if cont:
                continue  # move on to next time, since we don't have all the truth data

            # get HRRR data
            hrrr_pattern = dt.strftime(
                os.path.join(hrrrdir, "%Y%m%d", "HRRR_remapped_%Y%m%d_%H0000.nc*")
            )
            netcdf = glob.glob(hrrr_pattern)[0]
            if netcdf != curr_hrrr_file:
                curr_hrrr_file = netcdf  # reset
                hrrr_data = {
                    "mlcape": -1,
                    "u925": -1,
                    "v925": -1,
                    "u850": -1,
                    "v850": -1,
                    "u700": -1,
                    "v700": -1,
                    "u500": -1,
                    "v500": -1,
                }
                hrrr_status = read_netcdf(netcdf, data=hrrr_data)
                data["HRRR"] = hrrr_data

            # get SST data
            oisst_dt = dt - timedelta(days=1)  # help simulate realtime
            sst_pattern = oisst_dt.strftime(os.path.join(sstdir, "%Y%m%d*.nc"))
            # sst_pattern = oisst_dt.strftime('%Y%m%d*.nc')
            netcdf = glob.glob(sst_pattern)[0]
            if netcdf != curr_sst_file:
                curr_sst_file = netcdf  # reset
                sst_data = {"analysed_sst": -1, "lat": -1, "lon": -1}
                sst_status = read_netcdf(netcdf, data=sst_data)
                data["SST"] = sst_data

            # now go through the ABI grid (stride it) and collect patches
            datasets = {}
            for y in range(startY + halfsize + 1, endY, stride):
                for x in range(startX, endX - halfsize - 1, stride):
                    if landsea_mask[y, x] == 1:  # if water
                        # lat and lon
                        global_atts["lat"] = np.round(abi_lats[y, x], 5)
                        global_atts["lon"] = np.round(abi_lons[y, x], 5)
                        # time
                        global_atts["ABI_time"] = dt.strftime("%Y%m%d-%H%M%S UTC")

                        # ABI
                        for ch in abi_channels:
                            datasets["CH" + ch] = {}
                            if int(ch) == 2:  # 0.5km
                                yy = y * 4
                                xx = x * 4
                                hh = halfsize * 4
                                offset = 4
                                global_atts["ABIy05"] = yy
                                global_atts["ABIx05"] = int(xx)
                                patch = data["CH" + ch][
                                    yy - hh - offset : yy + hh + offset,
                                    xx - hh - offset : xx + hh + offset,
                                ]
                                if np.max(np.abs(patch)) > 2:
                                    cont = True  # QCing
                            elif int(ch) == 5 or int(ch) == 3 or int(ch) == 1:  # 1km
                                yy = y * 2
                                xx = x * 2
                                hh = halfsize * 2
                                offset = 2
                                global_atts["ABIy10"] = yy
                                global_atts["ABIx10"] = xx
                                patch = data["CH" + ch][
                                    yy - hh - offset : yy + hh + offset,
                                    xx - hh - offset : xx + hh + offset,
                                ]
                                if np.max(np.abs(patch)) > 2:
                                    cont = True  # QCing
                            else:  # 2km
                                yy = y
                                xx = x
                                hh = halfsize
                                offset = 1
                                global_atts["ABIy"] = yy
                                global_atts["ABIx"] = xx
                                patch = data["CH" + ch][
                                    yy - hh - offset : yy + hh + offset,
                                    xx - hh - offset : xx + hh + offset,
                                ]
                                if np.max(np.abs(patch)) > 500:
                                    cont = True  # QCing
                            datasets["CH" + ch]["data"] = patch

                        if cont == True:
                            continue  # go on to next (y,x)

                        # glm
                        offset = 1
                        for ch in glm_channels:
                            datasets[ch] = {
                                "data": data[ch][0][
                                    y - halfsize - offset : y + halfsize + offset,
                                    x - halfsize - offset : x + halfsize + offset,
                                ]
                            }
                        global_atts["maxFED_t0"] = np.max(
                            data["FED"][0][
                                y - GLM_truth_buffer : y + GLM_truth_buffer + 1,
                                x - GLM_truth_buffer : x + GLM_truth_buffer + 1,
                            ]
                        )
                        global_atts["maxFED_t5"] = np.max(
                            data["FED"][1][
                                y - GLM_truth_buffer : y + GLM_truth_buffer + 1,
                                x - GLM_truth_buffer : x + GLM_truth_buffer + 1,
                            ]
                        )
                        global_atts["maxFED_t10"] = np.max(
                            data["FED"][2][
                                y - GLM_truth_buffer : y + GLM_truth_buffer + 1,
                                x - GLM_truth_buffer : x + GLM_truth_buffer + 1,
                            ]
                        )
                        global_atts["maxFED_t15"] = np.max(
                            data["FED"][3][
                                y - GLM_truth_buffer : y + GLM_truth_buffer + 1,
                                x - GLM_truth_buffer : x + GLM_truth_buffer + 1,
                            ]
                        )
                        global_atts["maxFED_t20"] = np.max(
                            data["FED"][4][
                                y - GLM_truth_buffer : y + GLM_truth_buffer + 1,
                                x - GLM_truth_buffer : x + GLM_truth_buffer + 1,
                            ]
                        )
                        global_atts["maxFED_t25"] = np.max(
                            data["FED"][5][
                                y - GLM_truth_buffer : y + GLM_truth_buffer + 1,
                                x - GLM_truth_buffer : x + GLM_truth_buffer + 1,
                            ]
                        )
                        global_atts["maxFED_t30"] = np.max(
                            data["FED"][6][
                                y - GLM_truth_buffer : y + GLM_truth_buffer + 1,
                                x - GLM_truth_buffer : x + GLM_truth_buffer + 1,
                            ]
                        )

                        # hrrr
                        for ch in data["HRRR"]:
                            global_atts[ch] = data["HRRR"][ch][y, x]

                        # sst
                        # find correct pixel
                        sst_x = (np.abs(data["SST"]["lon"] - abi_lons[y, x])).argmin()
                        sst_y = (np.abs(data["SST"]["lat"] - abi_lats[y, x])).argmin()
                        global_atts["SST"] = np.round(
                            data["SST"]["analysed_sst"][0, sst_y, sst_x] - 273.15, 3
                        )  # convert to degrees C (has weird 3rd dim)

                        # write the final netCDFs for our database
                        dims = {}
                        dims["x"], dims["y"] = np.shape(datasets["CH13"]["data"])
                        dims["x05"], dims["y05"] = np.shape(datasets["CH02"]["data"])
                        dims["x10"], dims["y10"] = np.shape(datasets["CH05"]["data"])

                        output_file = os.path.join(
                            outbase,
                            yyyymmdd,
                            ("_").join(
                                [
                                    "goes16",
                                    "conus",
                                    "Y" + str(y),
                                    "X" + str(x),
                                    outts + ".netcdf",
                                ]
                            ),
                        )

                        for v in datasets:
                            datasets[v]["atts"] = vars_to_write[v][
                                "atts"
                            ]  # this sets the metadata for v
                            if v == "CH02":
                                currdim = ("x05", "y05")
                            elif v == "CH05":
                                currdim = ("x10", "y10")
                            elif v in [
                                "CH06",
                                "CH07",
                                "CH08",
                                "CH10",
                                "CH13",
                                "CH14",
                                "FED",
                                "FCD",
                                "TOE",
                                "AFA",
                            ]:
                                currdim = ("x", "y")
                            datasets[v][
                                "dims"
                            ] = currdim  # this sets the dimensions for v

                        utils.write_netcdf(
                            output_file, datasets, dims, atts=global_atts
                        )
                        # s = Popen(['gzip','-f',output_file]) #runs in background
                        # sys.exit()

                    # endif
                # endfor
            # endfor


if __name__ == "__main__":
    main()
