from netCDF4 import Dataset
import numpy as np

#########################################################################################################
def read_image_file(netcdf_file_name, NETCDF_PREDICTOR_NAMES):
    """Reads storm-centered images from NetCDF file.

    E = number of examples (storm objects) in file
    M = number of rows in each storm-centered grid
    N = number of columns in each storm-centered grid
    C = number of channels (predictor variables)

    :param netcdf_file_name: Path to input file.
    :return: image_dict: Dictionary with the following keys.
    image_dict['storm_ids']: length-E list of storm IDs (integers).
    image_dict['storm_steps']: length-E numpy array of storm steps (integers).
    image_dict['predictor_matrix']: E-by-M-by-N-by-C numpy array of predictor
        values.
    image_dict['target_name']: Name of target variable.
    image_dict['target_matrix']: E-by-M-by-N numpy array of target values.
    """

    dataset_object = Dataset(netcdf_file_name)

    predictor_matrix = None

    for this_predictor_name in NETCDF_PREDICTOR_NAMES:
        this_predictor_matrix = np.array(
            dataset_object.variables[this_predictor_name][:], dtype=float
        )
        # expand to 3 dims if we only have a slice and not a "cube"
        if this_predictor_matrix.ndim == 2:
            this_predictor_matrix = np.expand_dims(this_predictor_matrix, axis=0)
        # always expand the last "channel" dim
        this_predictor_matrix = np.expand_dims(this_predictor_matrix, axis=-1)

        if predictor_matrix is None:
            predictor_matrix = this_predictor_matrix + 0.0
        else:
            predictor_matrix = np.concatenate(
                (predictor_matrix, this_predictor_matrix), axis=-1
            )

    target_value = (
        np.zeros((predictor_matrix.shape[0]))
        if (dataset_object.severe_type == "non-severe")
        else np.ones((predictor_matrix.shape[0]))
    )

    dataset_object.close()

    return {"predictor_matrix": predictor_matrix, "target": target_value}


#########################################################################################################
def read_file_05km_2km(netcdf_file_name, NETCDF_PREDICTOR_NAMES):
    """Reads storm-centered images from NetCDF file.

    E = number of examples (storm objects) in file
    M = number of rows in each storm-centered grid
    N = number of columns in each storm-centered grid
    C = number of channels (predictor variables)

    :param netcdf_file_name: Path to input file.
    :return: image_dict: Dictionary with the following keys.
    image_dict['storm_ids']: length-E list of storm IDs (integers).
    image_dict['storm_steps']: length-E numpy array of storm steps (integers).
    image_dict['predictor_matrix']: E-by-M-by-N-by-C numpy array of predictor
        values.
    image_dict['target_name']: Name of target variable.
    image_dict['target_matrix']: E-by-M-by-N numpy array of target values.
    """

    dataset_object = Dataset(netcdf_file_name)

    predictor_matrix = None

    for this_predictor_name in NETCDF_PREDICTOR_NAMES:
        this_predictor_matrix = np.array(
            dataset_object.variables[this_predictor_name][:], dtype=float
        )
        # expand to 3 dims if we only have a slice and not a "cube"
        if this_predictor_matrix.ndim == 2:
            this_predictor_matrix = np.expand_dims(this_predictor_matrix, axis=0)
        # always expand the last "channel" dim
        this_predictor_matrix = np.expand_dims(this_predictor_matrix, axis=-1)

        if predictor_matrix is None:
            predictor_matrix = this_predictor_matrix + 0.0
        else:
            predictor_matrix = np.concatenate(
                (predictor_matrix, this_predictor_matrix), axis=-1
            )

    vis = np.array(dataset_object.variables["CH02"][:], dtype=float)

    target_value = (
        np.zeros((predictor_matrix.shape[0]))
        if (dataset_object.severe_type == "non-severe")
        else np.ones((predictor_matrix.shape[0]))
    )

    dataset_object.close()

    return {
        "predictor_matrix_2km": predictor_matrix,
        "vis": vis,
        "target": target_value,
    }
