import numpy as np
from subprocess import Popen, PIPE, call
import sys, os
import shapely as sp
from pyhdf.SD import SD, SDC
from scipy.spatial import ConvexHull
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
from datetime import datetime, timedelta
import time
import argparse
from netCDF4 import Dataset
from lightningcast import utils
import traceback

#################################################################################################################
def main():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", "--file", help="netCDF4 file containing data.", type=str, nargs=1
    )
    parser.add_argument(
        "-o", "--outdir", help="Output directory. Default = $PWD", type=str, nargs=1
    )
    parser.add_argument(
        "-l",
        "--logfile",
        help="logfile. Default = None.",
        default=["None"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-u", "--upload", help="Upload to RealEarth", action="store_true"
    )
    args = parser.parse_args()

    outdir = args.outdir[0] if (args.outdir) else os.environ["PWD"]
    logfile = None if (args.logfile[0] == "None") else args.logfile[0]

    # define the logger
    logDict = logging_def(logfile=logfile)
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl

    status = glm_objs_json(file=args.file[0], outdir=outdir, upload=args.upload)


##################################################################################################################
def glm_objs_json(file=None, data=None, outdir=os.environ["PWD"], upload=False):
    FUNCTION_NAME = "GLM_OBJS_JSON"

    logging.info(FUNCTION_NAME + ": Process started")

    utils.mkdir_p(outdir)

    if file:
        basefile = os.path.basename(file)
        yyyymmdd = basefile[0:8]
        hhmmss = basefile[9:15]
        logging.info(
            FUNCTION_NAME
            + ": Processing glm obj for "
            + yyyymmdd
            + "-"
            + hhmmss
            + " UTC"
        )

        # dt = datetime.strptime(yyyymmdd+hhmmss,'%Y%m%d%H%M%S')

        try:
            nc = Dataset(file, "r")
        except (OSError, IOError) as err:
            logging.error(FUNCTION_NAME + ": " + traceback.format_exc())
            return False
        ids = nc.variables["aid"][:]
        pids = nc.variables["pid"][:]
        data = nc.variables
        close_nc = True
    elif data:
        # data must be a dict
        yyyymmdd = data["timestamp"][0:8]
        hhmmss = data["timestamp"][9:15]
        ids = data["aid"]
        pids = data["pid"]
        close_nc = False
    else:
        logging.error(
            FUNCTION_NAME + ": You must supply either <FILE> or <DATA> argument"
        )
        return False

    # fvars = ['AID','CDATE','YEAR','MONTH','DAY','HOUR','FLASH_AREA','FLASH_ENERGY','LAT','LON','DURATION','NF1SIG','NF2SIG','NF3SIG',\
    #         'NGROUPS','MAXGRPENERGY','MEANGRPENERGY','MINGRPENERGY','MAXGRPDIS','NSERIES','MAXSERDUR','TOTSERDUR',\
    #          'NS1SIG','NS2SIG','NS3SIG','NEVENTS','MAXEVTENERGY','MINEVTENERGY','MAXEVTDIS']

    # avars = ['ID','ALAT','ALON','ATAI0','ATAI1','ANF2SIG','ANF2SIG','ANF3SIG','ANFLASHES','ANGROUPS','ANSERIES','ANS1SIG','ANS2SIG','ANS3SIG','ANEVENTS']

    # now compute create convex hull objs, compute some stats, and write geojson
    bopacity = "100"
    opacity = "20"
    width = "2"

    machine = os.uname()[1]
    currtime = datetime.utcnow().strftime("%Y%m%d_%H%M%S")
    js = open(
        ("/").join([outdir, "GLMOBJ_" + yyyymmdd + "_" + hhmmss + ".force.json"]), "w"
    )
    js.write(
        '{"source":"UW/CIMSS",\n "product":"GOES-16 GLM storm objects",\n "validTime":"'
        + yyyymmdd
        + "_"
        + hhmmss
        + ' UTC",\n "productionTime":"'
        + currtime
        + ' UTC",\
                  \n "machine":"'
        + machine
        + '",\n "type":"FeatureCollection",\n "features":[\n'
    )

    first_iter = True
    for ID in ids:
        good_area_ind = np.where(ids == ID)[0][0]
        good_flash_ind = np.where(pids == ID)
        points = np.array(
            [
                [x, y]
                for x, y in zip(
                    data["lon"][good_flash_ind], data["lat"][good_flash_ind]
                )
            ]
        )
        if len(points) >= 3:
            hull = ConvexHull(points)
            vertices = hull.vertices
            vertices = np.append(vertices, hull.vertices[0])
            coords = (
                "[[["
                + "],[".join(
                    "{:.3f},{:.3f}".format(lon, lat)
                    for lon, lat in zip(points[vertices, 0], points[vertices, 1])
                )
                + "]]]"
            )

            total_energy = str(int(np.sum(data["energy"][good_flash_ind]) * 1e15))
            nflashes = str(len(good_flash_ind[0]))
            ngroups = str(data["angroups"][good_area_ind])
            nevents = str(data["anevents"][good_area_ind])
            avg_area = str(int(np.mean(data["area"][good_flash_ind])))

            info = (
                "#flashes: "
                + nflashes
                + "<br>"
                + "#groups: "
                + ngroups
                + "<br>"
                + "#events: "
                + nevents
                + "<br>"
                + "total energy: "
                + total_energy
                + " fJ<br>"
                + "avg. area: "
                + avg_area
                + " km^2<br>"
                + "ID: "
                + str(ID)
            )

            if not (first_iter):
                js.write(",\n")  # complete previous feature block
            js.write('   {"type": "Feature",\n   "geometry": {\n')
            js.write(
                '     "type": "Polygon",\n     "coordinates":'
                + str(coords)
                + "\n   },\n"
            )
            js.write('   "properties": {\n')
            js.write('     "NFLASHES":"' + nflashes + '",\n')
            js.write('     "NGROUPS":"' + ngroups + '",\n')
            js.write('     "NEVENTS":"' + nevents + '",\n')
            js.write('     "AVG_AREA":"' + avg_area + '",\n')
            js.write('     "TOTAL_ENERGY":"' + total_energy + '",\n')
            js.write('     "INFO":"' + info + '",\n')
            js.write('     "WIDTH":"' + width + '",\n')
            js.write('     "OPACITY":"' + opacity + '",\n')
            js.write('     "BOPACITY":"' + bopacity + '",\n')
            js.write('     "COLOR":"0 0 0",\n')
            js.write('     "BCOLOR":"0 0 0"\n')
            js.write("   }}")
            if first_iter:
                first_iter = False

    js.write("]}")
    js.close()

    if close_nc:
        nc.close()

    logging.info(
        FUNCTION_NAME
        + ": Wrote out "
        + ("/").join([outdir, "GLMOBJ_" + yyyymmdd + "_" + hhmmss + ".force.json"])
    )

    if upload:
        status = Popen(
            [
                "./re_upload",
                os.path.join(
                    outdir, "GLMOBJ_" + yyyymmdd + "_" + hhmmss + ".force.json"
                ),
            ]
        )
    # print(points[hull.vertices,0]) #lons
    # print(points[hull.vertices,1]) #lats

    # import matplotlib.pyplot as plt
    # plt.plot(points[:,0], points[:,1], 'o')
    # for simplex in hull.simplices:
    #  plt.plot(points[simplex, 0], points[simplex, 1], 'k-')
    # plt.plot(points[hull.vertices,0], points[hull.vertices,1], 'r--', lw=2)
    # plt.plot(points[hull.vertices[0],0], points[hull.vertices[0],1], 'ro')
    # plt.show()

    return True


if __name__ == "__main__":
    main()
