import numpy as np
from netCDF4 import Dataset
from os import environ, makedirs, system
from os.path import isfile, isdir, basename
import os, sys
from subprocess import Popen, call
import time
from lightningcast import utils
from collections import OrderedDict
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
import argparse
import shutil
import traceback

###########################################################################################################################
def merge_scales():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--filename",
        help="filename only, (e.g., 20170508-174632.netcdf, or full .fml, if realtime)",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-dd",
        "--datadir",
        help="dirname for filename (e.g., /data/radar/ClusterID/).",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-o",
        "--outdir",
        help="outdir for final data (e.g., /data/prob/merged_objs/",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-v",
        "--var_name",
        help="variable name. Default = 'merged_prob_objs'",
        type=str,
        nargs=1,
        default=["merged_prob_objs"],
    )
    parser.add_argument("-l", "--logfile", help="logfile.", type=str, nargs=1)

    args = parser.parse_args()

    gzip = False

    filename = args.filename[0]

    # define the logger
    # this should handle 1) realtime, logfile set; 2) realtime no log set (uses default); 3) offline logfile set; 4) offline no logfile set (uses None)
    if args.logfile:
        logfile = args.logfile[0]
    else:
        logfile = None
    logDict = logging_def(logfile=logfile)
    # do this regardless of realtime/offline, so we get formatted stdout messages
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl
    ####################

    logging.info("Received " + filename)

    var_name = args.var_name[0]
    datadir = args.datadir[0]
    if datadir[-1] != "/":
        datadir += "/"
    if not (os.path.isdir(datadir)):
        logging.error(datadir + " does not exist or is not directory")
        sys.exit(1)
    outdir = args.outdir[0]
    if outdir[-1] != "/":
        outdir += "/"
    if not (isdir(outdir)):
        makedirs(outdir)

    # xind = np.zeros(10000000,dtype=int)
    # yind = np.zeros(10000000,dtype=int)

    metadata = {
        "LatGridSpacing": 1,
        "LonGridSpacing": 1,
        "Time": 1,
        "Latitude": 1,
        "Longitude": 1,
        "MissingData": 1,
    }
    scale_0 = utils.read_data(
        datadir + "scale_0/" + filename, "ClusterID", global_atts=metadata
    )
    scale_1 = utils.read_data(datadir + "scale_1/" + filename, "ClusterID")
    scale_2 = utils.read_data(datadir + "scale_2/" + filename, "ClusterID")

    ny, nx = np.shape(scale_0)

    merged_objs = scale_0  # start with scale_0 objs for final output grid

    scale_1_objs = np.unique(scale_1)
    scale_1_objs = scale_1_objs[1:]  # get rid of missing val

    # ==for each scale 1 object, get the spatial inds. Then find the unique scale_0 ids
    # for those spatial inds. If there is one positive ID and one missing value in the
    # unique vals [we want a missing value too, so that scale_1 obj and scale_0 obj don't
    # have the exact same spatial ind], then the merged_objs grid at the scale_1
    # spatial footprint is set to the scale_0 obj it encompasses.
    # ==Basically, if there is only one smaller scale object, spatially grow the object
    # to the bigger parent scale footprint. We do this at scale_2 as well, but with
    # the MERGED_OBJs -- not the scale_0 objs.
    logging.info("Marching through scale_1 to grow objects")
    for i in range(np.size(scale_1_objs)):
        # ind,ct = cython_where(scale_1,'EQ',scale_1_objs[i],xind,yind)
        ind = np.where(scale_1 == scale_1_objs[i])
        scale_0_objs = np.unique(scale_0[ind])  # ==scale_0 objs for scale_1 footprint
        if np.size(scale_0_objs) == 2 and scale_0_objs[0] <= 0:
            # ==missing data and 1 valid id
            merged_objs[ind] = scale_0_objs[1]

    # scale_2_objs = np.unique(scale_2)
    # scale_2_objs  = scale_2_objs[1:] #get rid of missing val

    # for i in range(np.size(scale_2_objs)):
    #  ind,ct = cython_where(scale_2,'EQ',scale_2_objs[i],xind,yind)
    #  ind = np.where(scale_2 == scale_2_objs[i])
    #  merged_objs = np.unique(merged_objs[ind])
    #  if(np.size(merged_objs) == 2 and merged_objs[0] <= 0):
    # ==missing data and 1 valid id
    #    merged_objs[ind] = merged_objs[1] #==grow the orig scale_0 obj to the scale_2 footprint

    netcdf = (
        outdir + filename[:-3] if (filename[-3:] == ".gz") else outdir + filename
    )  # remove .gz if necessary

    dims = OrderedDict()
    dims["Lat"] = ny
    dims["Lon"] = nx
    dataset = {
        var_name: {
            "data": merged_objs.astype(np.int32),
            "dims": ("Lat", "Lon"),
            "atts": {},
        }
    }  # note that int32 caps objectID at 2,147,483,647!
    atts = OrderedDict()
    atts["TypeName"] = var_name
    atts["DataType"] = "LatLonGrid"
    atts["LatGridSpacing"] = metadata["LatGridSpacing"]
    atts["LonGridSpacing"] = metadata["LonGridSpacing"]
    atts["Latitude"] = metadata["Latitude"]
    atts["Longitude"] = metadata["Longitude"]
    atts["Height"] = 0.0
    atts["Time"] = metadata["Time"]
    atts["FractionalTime"] = 0.0
    atts["attributes"] = "Unit ColorMap"
    atts["Unit-unit"] = "dimensionless"
    atts["Unit-value"] = "dimensionless"
    atts["ColorMap-unit"] = "dimensionless"
    atts["ColorMap-value"] = "KMeans"
    atts["MissingData"] = int(metadata["MissingData"])

    status = utils.write_netcdf(netcdf, dataset, dims, atts=atts, wait=True, gzip=gzip)


if __name__ == "__main__":
    merge_scales()
