import zarr
import time
import numpy as np
import xarray as xr
import pandas as pd
import os, sys
import pickle
from datetime import datetime, timedelta
import glob
import pyresample as pr

sys.path.insert(0, "..")
import utils
from netCDF4 import Dataset
import shutil
import argparse
from subprocess import Popen, call, PIPE

PWD = os.environ["PWD"] + "/"
import traceback
import solar_angles
import sat_zen_angle

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"  # visible in this process + all children

# CONSTANTS
# c = 2.99792458e8 #speed of light (m/s)
# h = 6.62607015e-34 #Planck's constant (J s)
# k = 1.38064852e-23 #Boltzmann's constant (J/K)
# l14 = 11.2e-6 #central wavelength of CH14 (m)

# formula for converting from radiance to BT.
# see https://www.star.nesdis.noaa.gov/goesr/docs/ATBD/Imagery.pdf
# T = [ fk2 / (alog((fk1 / Lλ) + 1)) - bc1 ] / bc2   eqn (3-5)

# for CH02 radiance to reflectance:
# see pdf above pp. 20-22
###############################################################################################################
def assign_data(netcdf, data, ch, PWD):

    baddata = False

    nc = Dataset(PWD + os.path.basename(netcdf), "r")
    radiance = nc.variables["Rad"][:]

    if int(ch) <= 6:
        # convert vis radiance to reflectance
        refl = radiance * nc.variables["kappa0"][:]
        data["CH" + ch] = refl
        if np.max(np.abs(refl)) > 2:
            baddata = True
    else:
        # convert IR radiance to brightness temperature
        bt = (
            nc.variables["planck_fk2"][:]
            / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
            - nc.variables["planck_bc1"][:]
        ) / nc.variables["planck_bc2"][:]
        data["CH" + ch] = bt
        if np.max(np.abs(bt)) > 500:
            baddata = True

    # close
    nc.close()

    return baddata


###############################################################################################################
def read_netcdf(file, data={}, global_atts={}):

    if file[-3:] == ".gz":
        gzip = 1
        unzipped_file = file[0 : len(file) - 3]
        the_file = unzipped_file
        try:
            s = call(["gunzip", "-f", file])
        except:
            print("cannot open " + file)
            return False
    else:
        gzip = 0
        the_file = file

    try:
        openFile = Dataset(the_file, "r")
    except (OSError, IOError):
        print(traceback.format_exc())
        return False

    for v in data:
        if v in openFile.variables:
            data[v] = openFile.variables[v][:]

    # get global attributes
    for gatt in global_atts:
        if gatt in openFile.ncattrs():
            global_atts[gatt] = getattr(
                openFile, gatt
            )  # global_atts is modified and returned
        else:
            print(gatt + " not in " + file)
            openFile.close()
            return False

    openFile.close()

    # --gzip file after closing?
    if gzip:
        s = Popen(
            ["gzip", "-f", unzipped_file]
        )  # ==will run in background since it doesn't need output from command.
        # s = call(['gzip','-f',unzipped_file])


###############################################################################################################
def get_glm_data(glm_file, data={}, global_atts={}, plax=None):

    if glm_file[-3:] == ".gz":
        gzip = 1
        unzipped_file = glm_file[0 : len(glm_file) - 3]
        the_file = unzipped_file
        try:
            s = call(["gunzip", "-f", glm_file])
        except:
            print("cannot open " + glm_file)
            return False
    else:
        gzip = 0
        the_file = glm_file

    try:
        openFile = Dataset(the_file, "r")
    except (OSError, IOError):
        print(traceback.format_exc())
        return False

    glmvars = [
        "flash_extent_density",
        "flash_centroid_density",
        "total_energy",
        "average_flash_area",
    ]
    for v in glmvars:
        if v in openFile.variables:
            if plax:  # this only works for 2D vars
                print("Parallax correcting " + v)
                tmpvar = openFile.variables[v][:]
                plax_var = np.zeros(np.shape(tmpvar))
                plax_var[plax] = np.ravel(tmpvar)
                data[v] = plax_var
            else:
                data[v] = openFile.variables[v][:]

    # get global attributes
    # for gatt in global_atts:
    #  if(gatt in openFile.ncattrs()):
    #    global_atts[gatt] = getattr(openFile,gatt) #global_atts is modified and returned
    #  else:
    #    print(gatt + ' not in ' + glm_file)
    #    openFile.close()
    #    return False

    openFile.close()

    # --gzip file after closing?
    if gzip:
        s = Popen(
            ["gzip", "-f", unzipped_file]
        )  # ==will run in background since it doesn't need output from command.
        # s = call(['gzip','-f',unzipped_file])

    return True


###############################################################################################################


def main():

    vars_to_write = {
        "CH02": {
            "atts": {
                "desc": "0.65-micron refl from ABI",
                "units": "dimensionless",
                "mean": 0.087134,
                "std": 0.15314,
            }
        },
        "CH05": {
            "atts": {
                "desc": "1.6-micron refl from ABI",
                "units": "dimensionless",
                "mean": 0.047332,
                "std": 0.0834376,
            }
        },
        "CH06": {
            "atts": {
                "desc": "2.26-micron refl from ABI",
                "units": "dimensionless",
                "mean": 0.04413,
                "std": 0.080274,
            }
        },
        "CH08": {
            "atts": {
                "desc": "6.2-micron BT from ABI",
                "units": "K",
                "mean": 234.73875,
                "std": 6.4855,
            }
        },
        "CH09": {"atts": {"desc": "6.9-micron BT from ABI", "units": "K"}},
        "CH10": {
            "atts": {
                "desc": "7.3-micron BT from ABI",
                "units": "K",
                "mean": 252.71758,
                "std": 11.06057,
            }
        },
        "CH11": {
            "atts": {
                "desc": "8.2-micron BT from ABI",
                "units": "K",
                "mean": 274.36105,
                "std": 21.69784,
            }
        },
        "CH13": {
            "atts": {
                "desc": "10.3-micron BT from ABI",
                "units": "K",
                "mean": 270.930,
                "std": 25.9126,
            }
        },
        "CH15": {
            "atts": {
                "desc": "12.4-micron BT from ABI",
                "units": "K",
                "mean": 273.168,
                "std": 19.9895,
            }
        },
        "CH16": {
            "atts": {
                "desc": "13.3-micron BT from ABI",
                "units": "K",
                "mean": 260.5381,
                "std": 17.2468,
            }
        },
        "FED": {
            "atts": {
                "desc": "GLM flash extent density, accumulated over 5 min",
                "units": "fl",
                "mean": 0.3536,
                "std": 2.341,
            }
        },
        "MAXFED_5to60": {
            "atts": {"desc": "max FED from 5 to 60 min in the future", "units": "fl"}
        },
        "MAXFED_5to45": {
            "atts": {"desc": "max FED from 5 to 45 min in the future", "units": "fl"}
        },
        "MAXFED_5to30": {
            "atts": {"desc": "max FED from 5 to 30 min in the future", "units": "fl"}
        },
        "satzen": {"atts": {"desc": "satellite-zenith angle"}},
        "solzen": {"atts": {"desc": "solar-zenith angle"}},
        "lat": {"atts": {"desc": "latitude"}},
        "lon": {"atts": {"desc": "longitude"}},
        "ABI_date": {"atts": {"desc": "ABI scan date (%Y%m%d), as an integer"}},
        "ABI_time": {"atts": {"desc": "ABI scan time (%H%M), as an integer"}},
        "startX": {
            "atts": {
                "desc": "The patch's nortwest X-index, in data space, from ABI CONUS 2-km fixed grid"
            }
        },
        "startY": {
            "atts": {
                "desc": "The patch's nortwest Y-index, in data space, from ABI CONUS 2-km fixed grid"
            }
        },
    }

    satdatadir = "/arcdata/goes/grb/goes16/"
    copy = True
    # sstdir = '/iliad/grain/common/oconus_ps/OISST/' #check GHRSST?
    ghrsst_dir = "/iliad/grain/common/oconus_ps/GHRSST_5km/"
    glmdir = "/apollo/grain/saved_probsevere_data/lightning/"
    enidir = glmdir
    hrrrdir = "/data/jcintineo/HRRR/remapped/"
    # outbase = '/apollo/grain/jcintineo/AI/OCONUS/seg/netcdf800x800/'
    zname = "/apollo/grain/jcintineo/AI/OCONUS/seg/unet_CONUS_800x800.zarr/"

    abi_channels = ["02", "05", "06", "08", "09", "10", "11", "13", "15", "16"]

    # get ABI lats/lons
    area_def, ny, nx, abi_lons, abi_lats = utils.get_area_definition(
        "../static/GOES_East.nc", return_latlons=True
    )
    # get land/sea mask
    mask_nc = Dataset("../static/mask_GOES-East_CONUS_25km.nc", "r")
    landsea_mask = mask_nc.variables["mask"][
        :
    ]  # this is on the ABI grid with effective resolution of 25km.
    mask_nc.close()

    predict_leadtime = 60  # in min
    pred_steps = predict_leadtime / 5 + 1

    ltg_frac_thresh = 0.005

    # with respect to GOES-East 2km ABI CONUS grid
    sX = [0, 800, 1600]
    sY = [0, 699]  # 699 because GLM 1499x2499 pre-202003
    npix = 800

    # sat_grid = pr.geometry.GridDefinition(lons=abi_lons, lats=abi_lats)
    # sat_proj_sst_grid = pr.geometry.GridDefinition(lons=abi_lons[::2,::2], lats=abi_lats[::2,::2]) #reducing to 4km res

    satzen = sat_zen_angle.sat_zen_angle(
        xlat=abi_lats, xlon=abi_lons, satlat=0, satlon=-75
    )

    data = {}
    data["satzen"] = satzen
    data["lat"] = abi_lats
    data["lon"] = abi_lons

    # important point about zarr syncronization
    # https://zarr.readthedocs.io/en/v2.1.1/tutorial.html?highlight=zip#parallel-computing-and-synchronization

    for d in list(range(20190726, 20190727)):

        strdate = str(d)
        topdt = datetime.strptime(strdate, "%Y%m%d")
        satdir1 = topdt.strftime(satdatadir + "/%Y/%Y_%m_%d_%j/abi/L1b/RadC/")

        satfiles = glob.glob(
            os.path.join(satdir1, topdt.strftime("OR_ABI-L1b-RadC-M*C13_G16_s%Y%j*"))
        )
        satfiles = np.sort(satfiles)
        satdts = np.array(
            [
                datetime.strptime(
                    os.path.basename(x).split("_s")[1][0:13], "%Y%j%H%M%S"
                )
                for x in satfiles
            ]
        )

        # for each satdt, get the t0, and t-sat_step grids
        for dt in satdts:
            cont = False
            yyyymmdd = dt.strftime("%Y%m%d")
            outts = dt.strftime("%Y%m%d-%H%M%S")

            # get all of the data first. Then do extraction.

            # first check if we have enough lightning to make it worth our while
            # get GLM data
            # we can't necessarily assume sat files will be every 5 min
            future_glm = []
            tmpdt = dt
            print("getting glm data")
            while len(future_glm) < pred_steps:
                netcdfpattern = tmpdt.strftime(
                    os.path.join(
                        glmdir, "%Y%m%d", "GLM", "goes_east", "agg", "%Y%m%d-%H%M*"
                    )
                )
                try:
                    netcdf = glob.glob(netcdfpattern)[0]
                except IndexError as err:
                    print(traceback.format_exc())
                    print(netcdfpattern)
                    cont = True
                    break
                else:
                    glm_data = {}
                    new_glm_data = get_glm_data(netcdf, data=glm_data)
                    future_glm.append(
                        glm_data["flash_extent_density"]
                    )  # [startY:endY,startX:endX])
                    tmpdt = tmpdt + timedelta(minutes=5)
            if cont:
                continue  # move on to next time, since we don't have all the truth data
            print("done")

            # solar zenith angle
            solzen, solaz = solar_angles.solar_angles(dt, lon=abi_lons, lat=abi_lats)
            data["solzen"] = solzen

            try:
                all_future_glm = np.stack(
                    future_glm[1:]
                )  # don't stack 0th time since it's ltg from t0-5 min to t0
            except ValueError:  # if exception, it's because some are (1499,2499) and some are (1500,2500). Just lose the last row/col.
                for ee in range(len(future_glm)):
                    future_glm[ee] = future_glm[ee][0:1499, 0:2499]
                all_future_glm = np.stack(future_glm[1:])

            max_future_glm = np.max(all_future_glm, axis=0)
            tot_pix = max_future_glm.shape[0] * max_future_glm.shape[1]
            good_ind = np.where(max_future_glm > 0)
            if len(good_ind[0]) > 0:
                ltg_frac = len(good_ind[0]) / tot_pix
            else:
                ltg_frac = 0

            if ltg_frac < ltg_frac_thresh:
                print(dt, "ltg_frac = ", np.round(ltg_frac, 6), "...not enough ltg")
                continue
            else:
                print(dt, "ltg_frac = ", np.round(ltg_frac, 6))
                max_glm_5to30 = np.max(all_future_glm[:6], axis=0)
                max_glm_5to45 = np.max(all_future_glm[:9], axis=0)

                # ltg frac is good, so get prev glm
                data["FED"] = future_glm[0]

                data["MAXFED_5to60"] = max_future_glm
                data["MAXFED_5to45"] = max_glm_5to45
                data["MAXFED_5to30"] = max_glm_5to30

            # get satellite data
            broke = False
            for ch in abi_channels:
                netcdfpattern = dt.strftime(
                    satdatadir
                    + "/%Y/%Y_%m_%d_%j/abi/L1b/RadC/OR_ABI-L1b-RadC-M*C"
                    + ch
                    + "_G16_s%Y%j%H%M*"
                )
                try:
                    netcdf = glob.glob(netcdfpattern)[0]
                except IndexError as err:
                    # a desired channel does not exist
                    print(traceback.format_exc())
                    print("Moving on to next sat DT...")
                    cont = True
                    break
                # copy file from satdatadir
                try:
                    shutil.copy(netcdf, PWD)
                except (PermissionError, OSError, IOError) as err:
                    print(traceback.format_exc())
                    print("Moving on to next sat DT...")
                    cont = True
                    break
                else:
                    baddata = assign_data(netcdf, data, ch, PWD)
                    os.remove(PWD + os.path.basename(netcdf))
                    if baddata:
                        cont = True
                        break

            if cont:
                continue  # continue to next satdt; can't get file or bad data

            # Now segment the CONUS grid into 800x800 2-km patches

            datasets = {}
            npatch = len(sX) * len(sY)
            datasets = {}
            datasets["ABI_time"] = {
                "data": np.zeros((npatch), dtype=int) + int(dt.strftime("%H%M"))
            }
            datasets["ABI_date"] = {
                "data": np.zeros((npatch), dtype=int) + int(dt.strftime("%Y%m%d"))
            }
            datasets["startX"] = {"data": np.zeros((npatch), dtype=int)}
            datasets["startY"] = {"data": np.zeros((npatch), dtype=int)}
            for key, val in data.items():
                cter = 0
                for (
                    sx
                ) in (
                    sX
                ):  # these are the startXs and startYs for the CONUS 2-km fixed grid
                    for sy in sY:
                        datasets["startX"]["data"][cter] = sx
                        datasets["startY"]["data"][cter] = sy
                        for key, val in data.items():
                            if len(val.shape) == 2:  # 2D
                                if key.startswith("CH"):  # ABI
                                    if key == "CH02":
                                        if cter == 0:
                                            datasets[key] = {
                                                "data": np.empty(
                                                    (npatch, npix * 4, npix * 4, 1),
                                                    dtype=np.float32,
                                                )
                                            }
                                        datasets[key]["data"][cter, ..., 0] = val[
                                            sy * 4 : sy * 4 + npix * 4,
                                            sx * 4 : sx * 4 + npix * 4,
                                        ]
                                    elif key in ["CH05", "CH03", "CH01"]:
                                        if cter == 0:
                                            datasets[key] = {
                                                "data": np.empty(
                                                    (npatch, npix * 2, npix * 2, 1),
                                                    dtype=np.float32,
                                                )
                                            }
                                        datasets[key]["data"][cter, ..., 0] = val[
                                            sy * 2 : sy * 2 + npix * 2,
                                            sx * 2 : sx * 2 + npix * 2,
                                        ]
                                    else:
                                        if cter == 0:
                                            datasets[key] = {
                                                "data": np.empty(
                                                    (npatch, npix, npix, 1),
                                                    dtype=np.float32,
                                                )
                                            }
                                        datasets[key]["data"][cter, ..., 0] = val[
                                            sy : sy + npix, sx : sx + npix
                                        ]
                                else:  # FED, MAXFED, satzen, solzen, etc.
                                    if cter == 0:
                                        datasets[key] = {
                                            "data": np.empty(
                                                (npatch, npix, npix, 1),
                                                dtype=np.float32,
                                            )
                                        }
                                    datasets[key]["data"][cter, ..., 0] = val[
                                        sy : sy + npix, sx : sx + npix
                                    ]

                        cter += 1

            # write to zarr
            ds_exists = True if (os.path.isdir(zname)) else False
            root = zarr.open(zname, mode="a")

            for key, val in datasets.items():
                # set chunksizes
                chunksize = 1
                if len(datasets[key]["data"].shape) == 1:
                    chunks = (npatch,)
                else:
                    chunks = (
                        chunksize,
                        datasets[key]["data"].shape[1],
                        datasets[key]["data"].shape[2],
                        1,
                    )

                # create data
                if ds_exists:  # just append along time axis
                    newarr = zarr.zeros(
                        datasets[key]["data"].shape,
                        chunks=chunks,
                        dtype=datasets[key]["data"].dtype,
                    )
                    newarr[:] = datasets[key]["data"]
                    root[key].append(newarr, axis=0)
                else:  # start of new data array in group
                    newarr = root.zeros(
                        key,
                        shape=datasets[key]["data"].shape,
                        chunks=chunks,
                        dtype=datasets[key]["data"].dtype,
                    )
                    newarr[:] = datasets[key]["data"]
                    # set attributes
                    for att in vars_to_write[key]["atts"]:
                        newarr.attrs[att] = vars_to_write[key]["atts"][att]

            # sys.exit()


if __name__ == "__main__":
    main()
