from __future__ import absolute_import, division, print_function, unicode_literals

import matplotlib

matplotlib.use("Agg")
import os

# os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID";
# os.environ["CUDA_VISIBLE_DEVICES"]="3";

import numpy as np
import pandas as pd
import xarray as xr
import zarr
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from lightningcast import performance_diagrams
from lightningcast import roc_curves
from lightningcast import attributes_diagrams
import pickle
from glob import glob
import copy

import tensorflow as tf

tf.keras.backend.clear_session()
from tensorflow import keras
import tensorflow.keras.backend as K

from lightningcast import tf_models
from tensorflow.keras.models import Model, save_model, load_model
from tensorflow.keras.callbacks import (
    EarlyStopping,
    ModelCheckpoint,
    ReduceLROnPlateau,
    CSVLogger,
)
from sklearn.metrics import mean_squared_error, roc_auc_score
from sklearn.metrics import auc as skl_auc
import sys
from lightningcast import utils
from datetime import datetime, timedelta
import time
import argparse

pltg = os.environ["PLTG"]
sys.path.insert(0, os.path.join(pltg, "src", "static/"))
from ABIcmaps import irw_ctc, wv_cimss_r
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
from lightningcast import keras_metrics
from lightningcast.LRFinder import LRFinder
from lightningcast import tf_eval

LIST_OF_METRIC_FUNCTIONS = [
    keras_metrics.accuracy,
    keras_metrics.binary_accuracy,
    keras_metrics.binary_csi,
    keras_metrics.binary_frequency_bias,
    keras_metrics.binary_pod,
    keras_metrics.binary_pofd,
    keras_metrics.binary_peirce_score,
    keras_metrics.binary_success_ratio,
    keras_metrics.binary_focn,
]
METRIC_FUNCTION_DICT = {
    "accuracy": keras_metrics.accuracy,
    "binary_accuracy": keras_metrics.binary_accuracy,
    "binary_csi": keras_metrics.binary_csi,
    "binary_frequency_bias": keras_metrics.binary_frequency_bias,
    "binary_pod": keras_metrics.binary_pod,
    "binary_pofd": keras_metrics.binary_pofd,
    "binary_peirce_score": keras_metrics.binary_peirce_score,
    "binary_success_ratio": keras_metrics.binary_success_ratio,
    "binary_focn": keras_metrics.binary_focn,
}

# METRIC_FUNCTION_DICT = {
#'accuracy': keras_metrics.accuracy,
#'iou':keras_metrics.iou,
#'iou_thresholded':keras_metrics.iou_thresholded
# }

##################################################################################################################
def training_history_figs(history, outdir):

    # summarize history for accuracy
    try:
        plt.plot(history["binary_accuracy"])
        if "val_binary_accuracy" in history:
            plt.plot(history["val_binary_accuracy"])
        plt.title("model binary accuracy")
        plt.ylabel("accuracy")
        plt.xlabel("epoch")
        if "val_binary_accuracy" in history:
            plt.legend(["train", "test"], loc="upper left")
        else:
            plt.legend(["train"], loc="upper left")
        plt.savefig(os.path.join(outdir, "binary_accuracy_history.png"))
        plt.close()
    except KeyError:
        plt.plot(history["accuracy"])
        if "val_accuracy" in history:
            plt.plot(history["val_accuracy"])
        plt.title("model accuracy")
        plt.ylabel("accuracy")
        plt.xlabel("epoch")
        if "val_accuracy" in history:
            plt.legend(["train", "test"], loc="upper left")
        else:
            plt.legend(["train"], loc="upper left")
        plt.savefig(os.path.join(outdir, "accuracy_history.png"))
        plt.close()

    # summarize history for loss
    plt.plot(history["loss"])
    if "val_loss" in history:
        plt.plot(history["val_loss"])
    plt.title("model loss")
    plt.ylabel("loss")
    plt.xlabel("epoch")
    if "val_loss" in history:
        plt.legend(["train", "test"], loc="upper left")
    else:
        plt.legend(["train"], loc="upper left")
    plt.savefig(os.path.join(outdir, "loss_history.png"))
    plt.close()


##################################################################################################################
def make_dists(sev_conv_pred, ns_conv_pred, outfig="dists.png"):

    fig = plt.figure(figsize=(7, 5), facecolor="white")
    ax1 = plt.gca()
    n1, bins1, patches1 = plt.hist(
        ns_conv_pred, 50, density=True, facecolor="b", alpha=0.25, stacked=True
    )
    n2, bins2, patches2 = plt.hist(
        sev_conv_pred, 50, density=True, facecolor="r", alpha=0.25, stacked=True
    )
    ax1.set_xlim(0, 0.7)
    ax1.set_ylabel("Conditional frequency")
    ax1.set_xlabel("CNN-predicted probability of severe storm")
    ax1.xaxis.grid(True, linestyle=":")

    ax2 = ax1.twinx()
    ratio = n2 / n1

    (ratio,) = plt.plot(bins1[0:50], ratio, color="green")
    ax2.set_ylabel("Sev/NS ratio")
    ax2.yaxis.label.set_color("green")
    ax2.set_ylim(0, 20)

    red_patch = mpatches.Patch(color="red", alpha=0.25, label="Severe storms")
    blue_patch = mpatches.Patch(color="blue", alpha=0.25, label="Non-severe storms")
    plt.legend(
        [blue_patch, red_patch, ratio],
        ["Non-severe storms", "Severe storms", "Sev/NS"],
        loc="upper center",
    )
    ax2.yaxis.grid(True, linestyle=":")

    fig.savefig(outfig, bbox_inches="tight")


##################################################################################################################


def plot_raw_img(
    bt_array_norm, bt_array, conv_graph, scaling=[180, 310], outfig=None, **kwargs
):

    conv_outs = conv_graph([bt_array_norm])

    fig = plt.figure(figsize=(3, 3), facecolor="white")
    num_conv_1 = conv_outs[0].shape[-1]
    ax_0 = fig.add_axes([0.01, 0.01, 0.98, 0.98], xticks=[], yticks=[])

    btimg = ax_0.pcolormesh(bt_array, vmin=scaling[0], vmax=scaling[1], cmap=cmap)
    plt.colorbar(btimg, ax=ax_0)
    ax_0.set_title(kwargs["title"], fontsize=12)
    ax_0.set_xlabel(
        f"Probability of severe storm: {conv_outs[-1][0, 0]:0.3f}", fontsize=12
    )

    if endcmd in kwargs:
        if kwargs["endcmd"] == "show":
            plt.show()

    if outfig:
        fig.savefig(outfig, bbox_inches="tight")

    plt.close()


##################################################################################################################


def plot_conv_layers_out(
    bt_array_norm, bt_array, conv_graph, scaling=[180, 310], outfig=None, **kwargs
):

    conv_outs = conv_graph([bt_array_norm])

    fig = plt.figure(figsize=(10, 10), facecolor="white")
    num_conv_1 = conv_outs[0].shape[-1]
    ax_0 = fig.add_axes([0.01, 0.6, 0.35, 0.35], xticks=[], yticks=[])

    btimg = ax_0.pcolormesh(bt_array, vmin=scaling[0], vmax=scaling[1], cmap=cmap)
    plt.colorbar(btimg, ax=ax_0)
    ax_0.set_title(kwargs["title"], fontsize=12)
    # ax_0.set_xlabel(f"Probability of severe storm: {conv_outs[-1][0, 0]:0.3f}", fontsize=12)
    ax_0.set_xlabel(
        "Probability of severe storm: " + kwargs["ex_ann"]["prob"], fontsize=12
    )

    if "ex_ann" in kwargs:
        fs = 12
        ex_ann = kwargs["ex_ann"]
        yyy = 0.55
        for k, v in ex_ann.items():
            if k == "prob":
                continue
            fig.text(0.01, yyy, k + ": " + v, fontsize=fs, ha="left", va="center")
            yyy -= 0.03

    bt_contours = [210, 230]
    for i in range(num_conv_1):
        ax_box = [
            0.38,
            i * 0.95 / num_conv_1 + 0.01,
            0.95 / num_conv_1,
            0.95 / num_conv_1,
        ]
        ax = fig.add_axes(ax_box, xticks=[], yticks=[])
        ax.pcolormesh(
            np.arange(0, 32, 2),
            np.arange(0, 32, 2),
            conv_outs[0][0, :, :, i],
            vmin=0,
            vmax=3,
            cmap="Reds",
        )
        ax.contour(
            np.arange(0, 31),
            np.arange(0, 31),
            bt_array,
            bt_contours,
            vmin=scaling[0],
            vmax=scaling[1],
            cmap="Blues_r",
        )
    ax.set_title("Conv. Layer 1", fontsize=fs)
    num_conv_2 = conv_outs[1].shape[-1]
    for i in range(num_conv_2):
        ax_box = [
            0.57,
            i * 0.95 / num_conv_2 + 0.01,
            0.95 / num_conv_2,
            0.95 / num_conv_2,
        ]
        ax = fig.add_axes(ax_box, xticks=[], yticks=[])
        ax.pcolormesh(
            np.arange(0, 32, 4),
            np.arange(0, 32, 4),
            conv_outs[1][0, :, :, i],
            vmin=0,
            vmax=3,
            cmap="Reds",
        )
        ax.contour(
            np.arange(0, 31),
            np.arange(0, 31),
            bt_array,
            bt_contours,
            vmin=scaling[0],
            vmax=scaling[1],
            cmap="Blues_r",
        )
    ax.set_title("Conv. Layer 2", fontsize=fs)

    num_conv_3 = conv_outs[2].shape[-1]
    for i in range(num_conv_3):
        width = 1 / num_conv_3 * 2
        if i < num_conv_3 // 2:
            ax_box = [0.69, i * 0.95 / num_conv_3 * 2 + 0.01, width, width]
        else:
            ax_box = [
                0.69 + width,
                0.95 * (i - num_conv_3 // 2) / num_conv_3 * 2 + 0.01,
                width,
                width,
            ]
        ax = fig.add_axes(ax_box, xticks=[], yticks=[])
        ax.pcolormesh(
            np.arange(0, 32, 8),
            np.arange(0, 32, 8),
            conv_outs[2][0, :, :, i],
            vmin=0,
            vmax=3,
            cmap="Reds",
        )
        ax.contour(
            np.arange(0, 31),
            np.arange(0, 31),
            bt_array,
            bt_contours,
            vmin=scaling[0],
            vmax=scaling[1],
            cmap="Blues_r",
        )
    fig.text(0.69 + width, 0.98, "Conv. Layer 3", fontsize=fs, ha="center", va="center")
    for i in range(num_conv_3):
        width = 1 / num_conv_3 * 2
        if i < num_conv_3 // 2:
            ax_box = [0.87, i * 0.95 / num_conv_3 * 2 + 0.01, width, width]
        else:
            ax_box = [
                0.87 + width,
                0.95 * (i - num_conv_3 // 2) / num_conv_3 * 2 + 0.01,
                width,
                width,
            ]
        ax = fig.add_axes(ax_box, xticks=[], yticks=[])
        ax.pcolormesh(
            np.arange(0, 32, 8),
            np.arange(0, 32, 8),
            conv_outs[2][0, :, :, i] * kwargs["final_weights"][:, :, i],
            vmin=-0.5,
            vmax=0.5,
            cmap="RdBu_r",
        )
        ax.contour(
            np.arange(0, 31),
            np.arange(0, 31),
            bt_array,
            bt_contours,
            vmin=scaling[0],
            vmax=scaling[1],
            cmap="Blues_r",
        )
    fig.text(0.87 + width, 0.98, "Weights", fontsize=12, ha="center", va="center")

    if endcmd in kwargs:
        if kwargs["endcmd"] == "show":
            plt.show()

    if outfig:
        fig.savefig(outfig, bbox_inches="tight")

    plt.close()


##################################################################################################################


def batch_imgs(
    df,
    inds,
    conv_graph,
    final_weights,
    outdir,
    plot_fct,
    endcmd="save",
    scaling=[180, 310],
    **kwargs,
):

    utils.mkdir_p(outdir)

    for idx in inds:
        outfig = os.path.join(outdir, str(idx) + ".png")

        ex_ann = {
            "ProbSevereID": str(df.iloc[idx]["ProbSevereID"]),
            "ABI_time": str(df.iloc[idx]["ABI_time"])[0:19],
            "mlat": "{:.2f}".format(df.iloc[idx]["mlat"]),
            "mlon": "{:.2f}".format(df.iloc[idx]["mlon"]),
            "prob": "{:.3f}".format(df.iloc[idx]["cnn_pred"]),
        }
        if df.iloc[idx]["hail_mags"] > 0:
            ex_ann["hail_mag"] = str(df.iloc[idx]["hail_mags"]) + " in"
            ex_ann["min_before_first_hail"] = (
                str(int(df.iloc[idx]["min_before_first_hail"])) + " min"
            )
            ex_ann["min_before_last_hail"] = (
                str(int(df.iloc[idx]["min_before_last_hail"])) + " min"
            )
        if df.iloc[idx]["wind_mags"] > 0:
            ex_ann["wind_mag"] = str(int(df.iloc[idx]["wind_mags"])) + " kts"
            ex_ann["min_before_first_wind"] = (
                str(int(df.iloc[idx]["min_before_first_wind"])) + " min"
            )
            ex_ann["min_before_last_wind"] = (
                str(int(df.iloc[idx]["min_before_last_wind"])) + " min"
            )
        if df.iloc[idx]["torn_mags"] >= 0:
            ex_ann["torn_mag"] = str(int(df.iloc[idx]["torn_mags"]))
            ex_ann["min_before_first_torn"] = (
                str(int(df.iloc[idx]["min_before_first_torn"])) + " min"
            )
            ex_ann["min_before_last_torn"] = (
                str(int(df.iloc[idx]["min_before_last_torn"])) + " min"
            )

        # quick normalize
        dat = df.iloc[idx]["ch14"]
        scaled_dat = (dat - scaling[0]) / (scaling[1] - scaling[0])

        scaled_dat = np.expand_dims(np.expand_dims(scaled_dat, axis=-1), axis=0)

        plot_fct(
            scaled_dat,
            dat,
            conv_graph,
            final_weights=final_weights,
            outfig=outfig,
            ex_ann=ex_ann,
            endcmd=endcmd,
            scaling=scaling,
            title=kwargs["title"],
        )


##################################################################################################################


def split_train_test(data, test_ratio, inds_only=False):
    """
    Use test_ratio to divide data into training and testing samples

    Args:
      data: a DataFrame or numpy array

      test_ratio: a fraction for testing sample. E.g., 0.2

      inds_only: A Boolean set whether or not to return only the indicies

    Returns:
      the training records and testing records (both are DataFrames)
    """

    shuffled_indices = np.random.permutation(len(data))
    test_set_size = int(len(data) * test_ratio)
    test_indices = shuffled_indices[:test_set_size]
    train_indices = shuffled_indices[test_set_size:]

    if inds_only:
        return train_indices, test_indices
    elif isinstance(data, pd.DataFrame):
        return data.iloc[train_indices], data.iloc[test_indices]
    elif isinstance(data, np.ndarray):
        return data[train_indices], data[test_indices]
    else:
        print("data are not DataFrame or ndarray")
        return -1, -1


##################################################################################################################


def normalize_multivariate_data(
    data, scaling_values=None, minmaxs=None, test=False, index=None, nframes=1
):
    """
    Normalize each channel in the 4 dimensional data matrix independently.

    Args:
        data: 4-dimensional array with dimensions (example, y, x, channel/variable)
        scaling_values: pandas dataframe containing mean and std columns
        minmaxs: A list of lists of max and min values used to scale each slice independently.
                These need to correctly correspond to each slice. If a the min and max for a slice
                are equal to -999, the mean and std will be used to normalize.
                E.g., [[180,300],[30,65],[0,2],[-999,-999]]

    Returns:
        normalized data array, scaling_values
    """
    normed_data = np.zeros(data.shape, dtype=data.dtype)
    scale_cols = (
        ["mean", "std"]
        if (not (minmaxs))
        else ["mean", "std", "min_scale", "max_scale"]
    )
    if scaling_values is None:
        scaling_values = pd.DataFrame(
            np.zeros((data.shape[-1], len(scale_cols)), dtype=np.float32),
            columns=scale_cols,
        )
    have_svs = False
    for i in range(data.shape[-1]):
        if not (test) and not (have_svs):
            scaling_values.loc[i, ["mean", "std"]] = [
                np.mean(data[..., i], dtype=np.float64),
                np.std(data[..., i], dtype=np.float64),
            ]
            if nframes > 1 and not (have_svs):
                # fill each "frame channel" scaling values with the 0th frame mean/std
                have_svs = True
                for ii in range(1, data.shape[-1]):
                    scaling_values.loc[ii, ["mean", "std"]] = scaling_values.loc[
                        0, ["mean", "std"]
                    ]

        if minmaxs:
            if minmaxs[i] != [-999, -999]:
                minval, maxval = minmaxs[i]
                dat = data[..., i]
                dat[dat > maxval] = maxval
                dat[dat < minval] = minval
                normed_data[..., i] = (dat - minval) / (maxval - minval)
                scaling_values.loc[i, ["min_scale", "max_scale"]] = [minval, maxval]
            else:
                if index:
                    normed_data[..., i] = (
                        data[..., i] - scaling_values.loc[index, "mean"]
                    ) / scaling_values.loc[index, "std"]
                    scaling_values.loc[index, ["min_scale", "max_scale"]] = [
                        minval,
                        maxval,
                    ]
                else:
                    normed_data[..., i] = (
                        data[..., i] - scaling_values.loc[i, "mean"]
                    ) / scaling_values.loc[i, "std"]
                    scaling_values.loc[i, ["min_scale", "max_scale"]] = [minval, maxval]
        else:
            if index:
                normed_data[..., i] = (
                    data[..., i] - scaling_values.loc[index, "mean"]
                ) / scaling_values.loc[index, "std"]
            else:
                normed_data[..., i] = (
                    data[..., i] - scaling_values.loc[i, "mean"]
                ) / scaling_values.loc[i, "std"]

    return normed_data, scaling_values


##################################################################################################################


def create_test_train(ABI_time, test_dates):

    # sev_recs = df.loc[df['severe_type'] == 1]
    # ns_recs = df.loc[df['severe_type'] == 0]
    # uniq_sev_psid = np.unique(sev_recs['ProbSevereID'])
    # uniq_ns_psid = np.unique(ns_recs['ProbSevereID'])

    # train_sev_ids, test_sev_ids = split_train_test(uniq_sev_psid,0.2)
    # train_ns_ids, test_ns_ids = split_train_test(uniq_ns_psid,0.2)

    dts = [
        datetime(int(str(td)[0:4]), int(str(td)[4:6]), int(str(td)[6:8]))
        for td in test_dates
    ]

    test_mask = np.isin(ABI_time, dts)
    test_index = ABI_time.loc[test_mask].index

    train_index = ABI_time.loc[~test_mask].index

    return train_index, test_index

    # train_recs, test_recs = split_train_test(df, 0.2)
    train_recs = train_recs.reset_index(drop=False)
    train_recs = train_recs.rename(index=int, columns={"index": "orig_index"})
    test_recs = test_recs.reset_index(drop=False)
    test_recs = test_recs.rename(index=int, columns={"index": "orig_index"})

    # get training patches
    logging.info("Getting training patches...")

    if len(channels) == 1:
        ch = channels[0]
        train_data = np.expand_dims(np.stack(list(train_recs[ch])), axis=-1)
    else:
        lists = list(zip(*[list(train_recs[ch]) for ch in channels]))
        train_data = np.transpose(
            np.stack(lists, axis=-1), (3, 2, 1, 0)
        )  # this is making the data in the right order. Shape of (n_examples, ny, nx, n_channels)
    train_out = np.array(train_recs["severe_type"])
    train_origIdx = np.array(train_recs["orig_index"])
    del train_recs

    # get test patches
    logging.info("Getting test patches...")

    if len(channels) == 1:
        ch = channels[0]
        test_data = np.expand_dims(
            np.stack(list(test_recs[ch])), axis=-1
        )  # shape of (n_examples, ny, nx, n_channels)
    else:
        lists = list(zip(*[list(test_recs[ch]) for ch in channels]))
        test_data = np.transpose(
            np.stack(lists, axis=-1), (3, 2, 1, 0)
        )  # this is making the data in the right order. Shape of (n_examples, ny, nx, n_channels)
        # test_data = np.stack([np.stack([i,j],axis=-1) for i,j in zip(list(test_recs[channels[0]]),list(test_recs[channels[1]]))]) #shape of (n_examples, ny, nx, n_channels)
        del lists
    test_out = np.array(test_recs["severe_type"])
    test_origIdx = np.array(test_recs["orig_index"])
    del test_recs

    # normalize the data
    logging.info("Normalizing data...")
    train_norm_2d, scaling_values = normalize_multivariate_data(
        train_data
    )  # ,minmaxs=[ch13_scaling])
    test_norm_2d, _ = normalize_multivariate_data(
        test_data, scaling_values=scaling_values, test=True
    )  # minmaxs=[ch13_scaling],scaling_values=scaling_values)

    return (
        train_norm_2d,
        train_out,
        train_origIdx,
        test_norm_2d,
        test_out,
        test_origIdx,
        scaling_values,
    )


##################################################################################################################
if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-m", "--model", help="import this trained CNN model", type=str, nargs=1
    )
    parser.add_argument(
        "-p",
        "--partition_file",
        help="Only valid if no --data (using a generator). Pickle file containing dict of train/val partition netcdfs and labels. Will also look for one in --outidr",
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-c",
        "--config_file",
        help="enter a config pkl file (usu. from a previous training session)",
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-l",
        "--logfile",
        help="Supply a logfile. Default is just to print to stdout",
        default=[None],
        nargs=1,
        type=str,
    )

    args = parser.parse_args()

    logDict = logging_def(logfile=args.logfile[0])
    # do this regardless of realtime/offline, so we get formatted stdout messages
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    # sys.stderr = sl

    if not (args.config_file):
        from lightningcast import tf_config

        config = tf_config.tf_config()
    else:
        config = pickle.load(open(args.config_file[0], "rb"))
    channels = config["channels"]
    scalar_vars = config["scalar_vars"]
    target = config["target"]
    batchsize = config["batchsize"]
    nframes = config["nframes"]
    binarize = config["binarize"]
    shuffle = config["shuffle"]

    from lightningcast import tf_generator

    DG = getattr(tf_generator, config["datagen"])

    outdir = config["outdir"]
    utils.mkdir_p(outdir)

    # print out config options
    ofile = os.path.join(outdir, "model_config.txt")
    of = open(ofile, "w")
    for key, value in config.items():
        of.write(str(key) + ": " + str(value) + "\n")

    pickle.dump(config, open(os.path.join(outdir, "model_config.pkl"), "wb"))

    ################################################
    csvlogger = CSVLogger(f"{outdir}/log.csv", append=True)
    early_stopping = EarlyStopping(monitor="val_loss", patience=config["es_patience"])
    mcp_save = ModelCheckpoint(
        os.path.join(outdir, "model-{epoch:02d}-{val_loss:03f}.h5"),
        save_best_only=True,
        monitor="val_loss",
        mode="min",
    )
    reduce_lr_loss = ReduceLROnPlateau(
        monitor="val_loss",
        cooldown=config["rlr_cooldown"],
        verbose=1,
        factor=config["rlr_factor"],
        patience=config["rlr_patience"],
        mode="min",
    )
    # lr_finder = LRFinder(min_lr=1e-7, max_lr=1e-1)
    # callbacks = [lr_finder]
    callbacks = [early_stopping, mcp_save, reduce_lr_loss, csvlogger]

    if len(config["data"]):  # if using a zarr file
        zf = zarr.open(config["data"])

        logging.info("Getting partition and labels data...")
        if args.partition_file:
            part_file = pickle.load(open(args.partition_file[0], "rb"))
            partition = part_file["partition"]
            labels = part_file["labels"]
            val_ind = partition["validation"]
            train_ind = partition["train"]
        elif os.path.isfile(os.path.join(outdir, "part_labs.pkl")):
            part_file = pickle.load(open(os.path.join(outdir, "part_labs.pkl"), "rb"))
            partition = part_file["partition"]
            labels = part_file["labels"]
            val_ind = partition["validation"]
            train_ind = partition["train"]
            train_mask = part_file["train_mask"]
            val_mask = part_file["val_mask"]
        else:
            partition = {}
            labels = {}

            days = zf.ABI_date[:]
            ind = np.arange(len(days))
            val_mask = np.isin(days, config["val_dates"])
            if len(config["train_dates"]):
                train_mask = np.isin(days, config["train_dates"])
            else:
                train_mask = np.invert(val_mask)
            val_ind = ind[val_mask]
            train_ind = ind[train_mask]

            partition["train"] = train_ind
            partition["validation"] = val_ind

            logging.info("Saving " + os.path.join(outdir, "part_labs.pkl..."))
            pickle.dump(
                {
                    "partition": partition,
                    "train_mask": train_mask,
                    "val_mask": val_mask,
                    "labels": labels,
                },
                open(os.path.join(outdir, "part_labs.pkl"), "wb"),
                protocol=4,
            )

        if config["datagen"] == "DataGeneratorLUSTREandZARR":
            # using both lustre and zarr samples
            train_list = pickle.load(open(config["train_list"], "rb"))
            training_generator = DG(train_list, train_ind, zf, config)
            val_list = pickle.load(open(config["val_list"], "rb"))
            validation_generator = DG(val_list, val_ind, zf, config)
            n_tsamples = len(train_list)
            n_vsamples = len(val_list)  # from lustre/samplelists only

            print(n_tsamples, n_vsamples)
            print(len(train_ind), len(val_ind))

            tmp_config = copy.deepcopy(config)
            tmp_config["shuffle"] = False
            tmp_config["batchsize"] = 1
            final_val_generator = DG(val_list, val_ind, zf, tmp_config)

    # Reading data directly from data partition, like /arcdata
    if config["datagen"] == "DataGeneratorLUSTRE":
        # We're using samplelists (pickles containing lists of dicts)
        train_list = pickle.load(open(config["train_list"], "rb"))
        training_generator = DG(
            train_list, config
        )  # , lats=latRR, lons=lonRR, satzen=satzenRR)
        val_list = pickle.load(open(config["val_list"], "rb"))
        validation_generator = DG(
            val_list, config
        )  # , lats=latRR, lons=lonRR, satzen=satzenRR)
        n_tsamples = len(train_list)
        n_vsamples = len(val_list)

        tmp_config = copy.deepcopy(config)
        tmp_config["shuffle"] = False
        tmp_config["batchsize"] = 1
        final_val_generator = DG(val_list, tmp_config)

    elif (
        config["datagen"] == "DataGeneratorUNET2D"
        or config["datagen"] == "DataGeneratorUNET3D"
    ):
        # Reading data batch from a zarr file
        training_generator = DG(partition["train"], zf, config)
        tmp_config = copy.deepcopy(config)
        tmp_config["shuffle"] = False
        validation_generator = DG(partition["validation"], zf, tmp_config)
        n_tsamples = len(partition["train"])
        n_vsamples = len(partition["validation"])

    # -----------------------------------------------------------------------------------------------------
    #  #extra generator stuff
    #  def t_gen():
    #    done = False
    #    iterator = iter(training_generator)
    #    while not done:
    #        try:
    #            yield next(iterator)
    #        except:
    #            done = True
    #
    #  def v_gen():
    #    done = False
    #    iterator = iter(validation_generator)
    #    while not done:
    #        try:
    #            yield next(iterator)
    #        except:
    #            done = True
    # # for i,y in v_gen():
    # #   print(i['input_1'].dtype,i['input_1'].shape,y.dtype)
    #
    # # output_types = tuple([tf.float32 for _ in range(len(config['inputs']))] + [tf.int32])
    # # output_shapes = []
    # # #one shape per input
    # # for inp in config['input_tuples']:
    # #   ny = inp[0]; nx = inp[1]; nchan = inp[2]
    # #   output_shapes.append(tf.TensorShape([batchsize, ny, nx, nchan]))
    # # #targets
    # # ny, nx = config['ps_2km']
    # # output_shapes.append(tf.TensorShape([batchsize, ny, nx]))
    # # output_shapes = tuple(output_shapes)
    #
    #
    #  output_shapes = ({'input_1':tf.TensorShape([batchsize, 1600, 1600, 1]),
    #                    'input_2':tf.TensorShape([batchsize, 800, 800, 1]),
    #                    'input_3':tf.TensorShape([batchsize, 400, 400, 1])}, tf.TensorShape([batchsize, 400, 400]))
    #
    #  output_types = ({'input_1':tf.float32,
    #                  'input_2':tf.float32,
    #                  'input_3':tf.float32}, tf.int32)
    #
    #  t_data = tf.data.Dataset.from_generator(
    #    t_gen,
    #    output_types=output_types,
    #    output_shapes=output_shapes
    #  )
    #
    #  v_data = tf.data.Dataset.from_generator(
    #    v_gen,
    #    output_types=output_types,
    #    output_shapes=output_shapes
    #  )
    #
    if config["ngpu"] > 1:
        # Create a MirroredStrategy.
        strategy = tf.distribute.MirroredStrategy()
        print("Number of devices: {}".format(strategy.num_replicas_in_sync))
        with strategy.scope():
            if args.model:
                logging.info("Opening " + args.model[0])
                conv_model = load_model(
                    args.model[0], compile=True
                )  # ,custom_objects=METRIC_FUNCTION_DICT)
                print(conv_model.summary())
                initial_epoch = int(args.model[0].split("-")[1])
            else:
                tf_model = getattr(tf_models, config["cnn"])
                conv_model = tf_model(config=config)
                initial_epoch = 0

        # t_data = strategy.experimental_distribute_dataset(t_data)
        # v_data = strategy.experimental_distribute_dataset(v_data)
    else:
        if args.model:
            logging.info("Opening " + args.model[0])
            conv_model = load_model(
                args.model[0], compile=True
            )  # ,custom_objects=METRIC_FUNCTION_DICT)
            print(conv_model.summary())
            initial_epoch = int(args.model[0].split("-")[1])
        else:
            tf_model = getattr(tf_models, config["cnn"])
            conv_model = tf_model(config=config)
            initial_epoch = 0
    #  #sys.exit()
    # -----------------------------------------------------------------------------------------------------
    logging.info("Fitting model with generator...")

    # FIT THE MODEL
    # shuffle has no effect if generator
    logging.info("Input shapes: " + str(config["input_tuples"]))
    #  history = conv_model.fit(t_data,validation_data=v_data,max_queue_size=10,
    #                             steps_per_epoch=len(train_ind)//batchsize, validation_steps=len(val_ind)//batchsize,
    #                             epochs=config['nepoch'],use_multiprocessing=True,
    #                             workers=config['nworkers'],verbose=0,callbacks=callbacks)
    history = conv_model.fit(
        training_generator,
        validation_data=validation_generator,
        max_queue_size=10,
        epochs=config["nepoch"],
        use_multiprocessing=True,
        initial_epoch=initial_epoch,
        workers=config["nworkers"],
        verbose=1,
        callbacks=callbacks,
    )

    logging.info("Saving model and training history...")
    save_model(conv_model, os.path.join(outdir, "fit_conv_model.h5"))
    pickle.dump(
        history.history,
        open(os.path.join(outdir, "training_history.pkl"), "wb"),
        protocol=4,
    )
    training_history_figs(history.history, outdir)

    ################################################
    # close config file
    of.close()

    # get predictions for validation data
    logging.info("Making predictions on validation data...")

    stride = 4
    if "unet" in config["datagen"].lower() or "lustre" in config["datagen"].lower():
        if config["datagen"] == "DataGeneratorLUSTREandZARR":
            # preallocate
            # for samplelists
            ny, nx = config["ps_2km"]
            preds = np.empty((n_vsamples, ny, nx), dtype=np.float16)
            val_target = np.empty((n_vsamples, ny, nx), dtype=np.float16)
            # for zarr
            ny = nx = 400  # hard coding
            zshape = (1, ny, nx)  # hard coding
            zarr_preds = np.empty((len(val_ind), ny, nx), dtype=np.float16)
            zarr_target = np.empty((len(val_ind), ny, nx), dtype=np.float16)

            cter1 = cter2 = 0
            for inputs, labels in final_val_generator:
                tmp = conv_model.predict(inputs, workers=10, verbose=1)
                if tmp.shape == zshape:  # from zarr
                    zarr_preds[cter1] = tmp
                    zarr_target[cter1] = labels
                    cter1 += 1
                else:
                    preds[cter2] = tmp
                    val_target[cter2] = labels
                    cter2 += 1

            preds = preds[::stride, ::stride].flatten()
            val_target = val_target[::stride, ::stride].flatten()
            zarr_preds = zarr_preds[::stride, ::stride].flatten()
            zarr_target = zarr_target[::stride, ::stride].flatten()
            preds = np.concatenate((preds, zarr_preds)).astype(
                np.float32
            )  # float32 precision needed for validation routines (e.g., MSE)
            val_target = np.concatenate((val_target, zarr_target)).astype(
                np.float32
            )  # float32 precision needed for validation routines (e.g., MSE)

        elif (
            config["datagen"] == "DataGeneratorLUSTRE"
            or config["datagen"] == "DataGeneratorUNET2D"
            or config["datagen"] == "DataGeneratorUNET3D"
        ):
            # preallocate
            ny, nx = config["ps_2km"]
            preds = np.empty((n_vsamples, ny, nx), dtype=np.float16)
            val_target = np.empty((n_vsamples, ny, nx), dtype=np.float16)

            if config["spatial_verification"]:
                thresholds = np.array(
                    [0.1, 0.2, 0.3, 0.35, 0.4, 0.5, 0.6, 0.7, 0.8, 0.90]
                )
                all_hits, all_misses, all_FAs = [
                    np.zeros((len(thresholds), 1500, 2500), dtype=np.int32)
                    for _ in range(3)
                ]  # hard-coded for ABI CONUS

            cter = 0
            for (
                inputs,
                labels,
            ) in (
                final_val_generator
            ):  # shuffle should be False, especially if spatial_verification == True
                preds[cter] = conv_model.predict(inputs, workers=10, verbose=1)
                val_target[cter] = labels

                if config["spatial_verification"]:
                    startX = val_list[cter]["startX"]
                    endX = val_list[cter]["endX"]
                    startY = val_list[cter]["startY"]
                    endY = val_list[cter]["endY"]
                    hits, misses, FAs = tf_eval.spatial_verification(
                        preds=preds[cter],
                        targets=val_target[cter],
                        thresholds=thresholds,
                        target_thresh=config["binarize"],
                    )
                    # add to final
                    all_hits[:, startY:endY, startX:endX] += hits
                    all_misses[:, startY:endY, startX:endX] += misses
                    all_FAs[:, startY:endY, startX:endX] += FAs

                cter += 1

            preds = (
                preds[::stride, ::stride].flatten().astype(np.float32)
            )  # float32 precision needed for validation routines (e.g., MSE)
            val_target = (
                val_target[::stride, ::stride].flatten().astype(np.float32)
            )  # float32 precision needed for validation routines (e.g., MSE)

        else:
            print("not implemented")
            sys.exit()
            val_target = zf[target].get_orthogonal_selection(
                (val_ind, slice(None), slice(None), [-1])
            )
            preds = (
                preds[::stride, ::stride].flatten().astype(np.float32)
            )  # float32 precision needed for validation routines (e.g., MSE)
            val_target = (
                val_target[::stride, ::stride].flatten().astype(np.float32)
            )  # float32 precision needed for validation routines (e.g., MSE)

        val_target[val_target < binarize] = 0
        val_target[val_target >= binarize] = 1
        assert len(preds) == len(val_target)

    else:
        preds = preds.reshape(len(preds))
        del val_generator

        logging.info(
            "Saving preds and val_target to "
            + os.path.join(outdir, "cnn_pred_labs.pkl")
        )
        print("len(val_target):", len(val_target), "len(preds):", len(preds))
        val_target = val_target[0 : len(preds)]
        pred_df = pd.DataFrame(
            np.column_stack((preds, val_target)), columns=["cnn_pred", "labels"]
        )
        pred_df.to_pickle(os.path.join(outdir, "cnn_pred_labs.pkl"))
        del pred_df

    #####################################################################################################
    #####################################################################################################

    scores_dict = tf_eval.verification(preds, val_target, outdir)
    pickle.dump(scores_dict, open(os.path.join(outdir, "scores_dict.pkl"), "wb"))

    # spatial verification file and images
    if config["spatial_verification"]:
        datasets = {}
        dims = {"t": len(thresholds), "y": 1500, "x": 2500}
        datasets["hits"] = {"data": all_hits, "dims": ("t", "y", "x"), "atts": {}}
        datasets["misses"] = {"data": all_misses, "dims": ("t", "y", "x"), "atts": {}}
        datasets["FAs"] = {"data": all_FAs, "dims": ("t", "y", "x"), "atts": {}}
        datasets["thresholds"] = {"data": thresholds, "dims": ("t"), "atts": {}}
        global_atts = {"samplelist": config["val_list"]}
        global_atts["sample_shape"] = f"({n_vsamples},{ny},{nx})"
        global_atts["created"] = datetime.utcnow().strftime("%Y%m%d-%H%M UTC")
        utils.write_netcdf(
            f"{outdir}/spatial_counts.nc", datasets, dims, atts=global_atts, gzip=False
        )

        from lightningcast import spatial_csi

        spatial_csi.main(f"{outdir}/spatial_counts.nc", 0.35)
